package de.fhg.fokus.edp.chatbot_sparql_service;

import com.fasterxml.jackson.databind.MapperFeature;
import de.fhg.fokus.edp.chatbot_sparql_service.model.Dataset;
import de.fhg.fokus.edp.chatbot_sparql_service.model.DatasetRequest;
import de.fhg.fokus.edp.chatbot_sparql_service.model.Intent;
import de.fhg.fokus.edp.chatbot_sparql_service.model.Mode;
import io.vertx.config.ConfigRetriever;
import io.vertx.core.AbstractVerticle;
import io.vertx.core.CompositeFuture;
import io.vertx.core.DeploymentOptions;
import io.vertx.core.Future;
import io.vertx.core.eventbus.DeliveryOptions;
import io.vertx.core.http.HttpServer;
import io.vertx.core.http.HttpServerOptions;
import io.vertx.core.json.DecodeException;
import io.vertx.core.json.Json;
import io.vertx.core.json.JsonArray;
import io.vertx.core.json.JsonObject;
import io.vertx.ext.web.Router;
import io.vertx.ext.web.RoutingContext;
import io.vertx.ext.web.api.contract.RouterFactoryOptions;
import io.vertx.ext.web.api.contract.openapi3.OpenAPI3RouterFactory;
import io.vertx.ext.web.handler.StaticHandler;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;

import static de.fhg.fokus.edp.chatbot_sparql_service.ApplicationConfig.*;


public class MainVerticle extends AbstractVerticle {

    private static final Logger LOG = LoggerFactory.getLogger(MainVerticle.class);

    private static final Map<String, List<Dataset>> storedDatasets = new ConcurrentHashMap<>();

    private JsonObject config;

    @Override
    public void start(Future<Void> completionFuture) {
        LOG.info("Launching Chatbot-SPARQL-Service...");

        // startup is only successful if no step failed
        Future<Void> steps = loadConfig()
            .compose(handler -> bootstrapVerticles())
            .compose(handler -> startServer());

        steps.setHandler(handler -> {
            if (handler.succeeded()) {
                LOG.info("Chatbot-SPARQL-Service successfully launched");
                completionFuture.complete();
            } else {
                LOG.error("Failed to launch Chatbot-SPARQL-Service: " + handler.cause());
                completionFuture.fail(handler.cause());
            }
        });
    }

    private Future<Void> loadConfig() {
        Future<Void> future = Future.future();

        ConfigRetriever configRetriever = ConfigRetriever.create(vertx);

        configRetriever.getConfig(handler -> {
            if (handler.succeeded()) {
                config = handler.result();
                LOG.debug(config.encodePrettily());
                future.complete();
            } else {
                future.fail("Failed to load config: " + handler.cause());
            }
        });

        configRetriever.listen(change ->
            config = change.getNewConfiguration());

        // accept case insensitive enums
        Json.mapper.enable(MapperFeature.ACCEPT_CASE_INSENSITIVE_ENUMS);

        return future;
    }

    private Future bootstrapVerticles() {
        DeploymentOptions options = new DeploymentOptions()
            .setConfig(config)
            .setWorker(true);

        List<Future> deploymentFutures = new ArrayList<>();
        deploymentFutures.add(startVerticle(options, SparqlVerticle.class.getName()));

        return CompositeFuture.all(deploymentFutures);
    }

    private Future<Void> startServer() {
        Future<Void> startFuture = Future.future();
        Integer port = config.getInteger(ENV_APPLICATION_PORT, DEFAULT_APPLICATION_PORT);

        OpenAPI3RouterFactory.create(vertx, "webroot/openapi.yaml", handler -> {
            if (handler.succeeded()) {
                OpenAPI3RouterFactory routerFactory = handler.result();
                RouterFactoryOptions options = new RouterFactoryOptions().setMountNotImplementedHandler(true).setMountValidationFailureHandler(true);
                routerFactory.setOptions(options);

                routerFactory.addHandlerByOperationId("retrieveDatasets", this::handleDatasetRequest);

                Router router = routerFactory.getRouter();
                router.route("/*").handler(StaticHandler.create());

                HttpServer server = vertx.createHttpServer(new HttpServerOptions().setPort(port));
                server.requestHandler(router).listen();

                LOG.info("Server successfully launched on port [{}]", port);
                startFuture.complete();
            } else {
                // Something went wrong during router factory initialization
                LOG.error("Failed to start server at [{}]: {}", port, handler.cause());
                startFuture.fail(handler.cause());
            }
        });

        return startFuture;
    }

    private void handleDatasetRequest(RoutingContext context) {
        LOG.info("Received request: [{}]", context.getBodyAsString());

        try {
            // parse incoming body once to check if request is valid
            DatasetRequest datasetRequest = Json.decodeValue(context.getBody(), DatasetRequest.class);

            Intent intent = Intent.valueOf(datasetRequest.getIntent() != null ? datasetRequest.getIntent().toUpperCase() : Intent.NEW_SEARCH.name());
            Mode mode = Mode.valueOf(datasetRequest.getMode() != null ? datasetRequest.getMode().toUpperCase() : DEFAULT_MODE.name());
            int pageSize = datasetRequest.getPageSize() != null ? datasetRequest.getPageSize() : config.getInteger(ENV_PAGE_SIZE, DEFAULT_PAGE_SIZE);

            context.response().putHeader("Content-Type", "application/json");

            switch (intent) {

                case NEW_SEARCH:
                    performNewSearch(context, datasetRequest.getBotId(), mode, pageSize);
                    break;

                case NEXT_RESULT_PAGE:
                    getNextResultPage(context, datasetRequest.getBotId(), mode, datasetRequest.getOffset(), pageSize);
                    break;
            }
        } catch (DecodeException e) {
            context.response().setStatusCode(400).end(e.getMessage());
        }
    }

    private Future<Void> startVerticle(DeploymentOptions options, String className) {
        Future<Void> future = Future.future();

        vertx.deployVerticle(className, options, handler -> {
            if (handler.succeeded()) {
                future.complete();
            } else {
                LOG.error("Failed to deploy verticle [{}] : {}", className, handler.cause());
                future.fail("Failed to deploy [" + className + "] : " + handler.cause());
            }
        });

        return future;
    }

    private void performNewSearch(RoutingContext context, String botId, Mode mode, int pageSize) {
        vertx.eventBus().send(SPARQL_VERTICLE_ADDRESS, context.getBodyAsString(), new DeliveryOptions().setSendTimeout(120000), sendHandler -> {
            if (sendHandler.succeeded()) {
                JsonObject result = ((JsonObject) sendHandler.result().body());
                if (result.getBoolean("success")) {

                    List<Dataset> results = Arrays.asList(Json.decodeValue(result.getJsonArray("result").toString(), Dataset[].class));
                    storedDatasets.put(botId, results);

                    context.response()
                        .setStatusCode(200)
                        .end(formatResults(results, mode, 0, pageSize));
                } else {
                    context.response()
                        .setStatusCode(500)
                        .end(result.getString("message"));
                }
            } else {
                LOG.error("Failed to handle dataset retrieval info request: [{}]", sendHandler.cause().getMessage());

                JsonObject response = new JsonObject();
                response.put("message", "Failed to handle dataset retrieval request: " + sendHandler.cause());

                context.response().setStatusCode(500).end(response.encode());
            }
        });
    }

    private void getNextResultPage(RoutingContext context, String botId, Mode mode, int offset, int pageSize) {
        if (storedDatasets.containsKey(botId)) {
            context.response()
                .setStatusCode(200)
                .end(formatResults(storedDatasets.get(botId), mode, offset, pageSize));
        } else {
            context.response()
                .setStatusCode(200)
                .end(new JsonObject().put("success", false).encodePrettily());
        }
    }

    private String formatResults(List<Dataset> datasets, Mode mode, int offset, int pageSize) {

        String result = null;

        switch (mode) {
            case STANDARD:
                result = new JsonObject()
                    .put("success", true)
                    .put("totalCount", datasets.size())
                    .put("result", datasets)
                    .encodePrettily();
                break;

            case FORMATTED:
                if (offset >= 0 && offset < datasets.size()) {
                    StringBuilder datasetBuilder = new StringBuilder();

                    int toIndex = Math.min(offset + pageSize, datasets.size());

                    datasets.subList(offset, toIndex).forEach(dataset ->
                        datasetBuilder.append(dataset.humanReadable()).append("\n\n"));

                    result = new JsonObject()
                        .put("success", true)
                        .put("totalCount", datasets.size())
                        .put("offset", offset)
                        .put("pageSize", pageSize)
                        .put("result", datasetBuilder.toString())
                        .encodePrettily();
                } else {
                    result = new JsonObject()
                        .put("success", false)
                        .encodePrettily();
                }
                break;

            case WOO:
                result = datasets.stream().map(Dataset::humanReadable).collect(Collectors.joining("\n\n"));
                break;
        }

        LOG.debug("Formatted results: {}", result);
        return result;
    }
}
