package de.fhg.fokus.edp.chatbot_sparql_service;

import de.fhg.fokus.edp.chatbot_sparql_service.model.DateConstraint;
import de.fhg.fokus.edp.chatbot_sparql_service.model.rdf.AUTHORITY;
import org.apache.commons.lang3.StringUtils;
import org.apache.jena.atlas.lib.StrUtils;
import org.apache.jena.datatypes.xsd.XSDDatatype;
import org.apache.jena.graph.NodeFactory;
import org.apache.jena.rdf.model.Literal;
import org.apache.jena.rdf.model.Model;
import org.apache.jena.rdf.model.Resource;
import org.apache.jena.sparql.core.Var;
import org.apache.jena.sparql.expr.*;
import org.apache.jena.sparql.expr.nodevalue.NodeFunctions;
import org.apache.jena.sparql.expr.nodevalue.NodeValueDT;
import org.apache.jena.sparql.expr.nodevalue.NodeValueString;
import org.apache.jena.vocabulary.DC;
import org.apache.jena.vocabulary.SKOS;

import java.time.*;
import java.time.format.DateTimeFormatter;
import java.time.temporal.TemporalAdjusters;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

public class SparqlUtils {

    /**
     * Recursively combines all provided Jena expressions using conjunction (logical AND)
     *
     * @param expressions List of expressions
     * @return An expression consisting of disjunctions (or the expression itself, if a singleton list is provided)
     */
    static Expr recursiveConjunction(List<Expr> expressions) {
        if (expressions.size() == 1)
            return expressions.get(0);

        return expressions.size() == 2
            ? new E_LogicalAnd(expressions.get(0), expressions.get(1))
            : new E_LogicalAnd(expressions.remove(expressions.size() - 1), recursiveConjunction(expressions));
    }

    /**
     * Recursively combines all provided Jena expressions using disjunction (logical OR)
     *
     * @param expressions List of expressions
     * @return An expression consisting of disjunctions (or the expression itself, if a singleton list is provided)
     */
    static Expr recursiveDisjunction(List<Expr> expressions) {
        if (expressions.size() == 1)
            return expressions.get(0);

        return expressions.size() == 2
            ? new E_LogicalOr(expressions.get(0), expressions.get(1))
            : new E_LogicalOr(expressions.remove(expressions.size() - 1), recursiveDisjunction(expressions));
    }

    /**
     * Creates a filter for the language tag of a given variable.
     * Currently only supports english. In the EDP an empty language tag indicates the title being natively english,
     * whilst a string of the form en-* indicates an english translation.
     *
     * @param var The variable to match the language tag on
     * @return An Expression containing a filter
     */
    static Expr getLanguageFilter(Var var) {
        return new E_LogicalOr(
            new E_Equals(new E_Lang(new ExprVar(var)), new NodeValueString("")),
            new E_StrStartsWith(new E_Lang(new ExprVar(var)), new NodeValueString("en")));
    }


    /**
     * Parses a given string into a {@link Resource} that resembles the input.
     * The properties matched are SKOS.prefLabel, DC.identifier and AUTHORITY.opMappedCode, the latter being a property of the class AUTHORITY.legacyCode.
     * The resource must be contained in the model passed. If there is more than one match a random one is returned.
     *
     * @param model The model containing all resources which are checked for matching property values
     * @param userLanguage The input used for matching
     * @return An optional containing a resource if one has been found, else an empty optional
     */
    static Optional<Resource> getLinguisticSystemByLanguageName(Model model, String userLanguage) {

        if (userLanguage == null || userLanguage.isEmpty())
            return Optional.empty();

        // try to match prefLabel (eg English)
        Literal prefLabel = model.createLiteral(StringUtils.capitalize(userLanguage.toLowerCase()), "en");
        Set<Resource> linguisticSystems = new HashSet<>(model.listResourcesWithProperty(SKOS.prefLabel, prefLabel).toSet());

        // try to match authority code (eg ENG)
        if (linguisticSystems.isEmpty())
            linguisticSystems.addAll(model.listResourcesWithProperty(DC.identifier, StringUtils.capitalize(userLanguage.toUpperCase())).toSet());

        // try to match legacy code (eg EN)
        if (linguisticSystems.isEmpty())
            model.listResourcesWithProperty(AUTHORITY.legacyCode, userLanguage.toUpperCase()).toSet()
                .forEach(mappedCode -> linguisticSystems.addAll(model.listResourcesWithProperty(AUTHORITY.opMappedCode, mappedCode).toSet()));

        return linguisticSystems.isEmpty()
            ? Optional.empty()
            : Optional.of(linguisticSystems.iterator().next());
    }

    /**
     * Parses a given string into a {@link Resource} that resembles the input.
     * The properties matched are SKOS.prefLabel, DC.identifier and AUTHORITY.authorityCode.
     * The resource must be contained in the model passed. If there is more than one match a random one is returned.
     *
     * @param model The model containing all resources which are checked for matching property values
     * @param format The input used for matching
     * @return An optional containing a resource if one has been found, else an empty optional
     */
    static Optional<Resource> getMediaTypeByFormatName(Model model, String format) {

        if (format == null || format.isEmpty())
            return Optional.empty();

        // try to match prefLabel
        Literal prefLabel = model.createLiteral(format.toUpperCase(), "en");
        Set<Resource> mediaTypes = new HashSet<>(model.listResourcesWithProperty(SKOS.prefLabel, prefLabel).toSet());

        // try to match identifier
        if (mediaTypes.isEmpty())
            mediaTypes.addAll(model.listResourcesWithProperty(DC.identifier, format.toUpperCase()).toSet());

        // try to match authority code
        if (mediaTypes.isEmpty())
            mediaTypes.addAll(model.listResourcesWithProperty(AUTHORITY.authorityCode, format.toUpperCase()).toSet());

        return mediaTypes.isEmpty()
            ? Optional.empty()
            : Optional.of(mediaTypes.iterator().next());
    }

    /**
     * Parses a given string into a {@link Resource} that resembles the input.
     * The properties matched are SKOS.prefLabel, DC.identifier and AUTHORITY.authorityCode.
     * The resource must be contained in the model passed. If there is more than one match a random one is returned.
     *
     * @param model The model containing all resources which are checked for matching property values
     * @param format The input used for matching
     * @return An optional containing a resource if one has been found, else an empty optional
     */
    static Optional<Resource> getLocationByName(Model model, String location) {

        if (location == null || location.isEmpty())
            return Optional.empty();

        // try to match prefLabel
        String capitalized = Arrays.stream(location.split("\\s+")).map(loc -> StringUtils.capitalize(loc.toLowerCase())).collect(Collectors.joining(" "));
        Literal prefLabel = model.createLiteral(capitalized, "en");
        Set<Resource> locations = new HashSet<>(model.listResourcesWithProperty(SKOS.prefLabel, prefLabel).toSet());

        // try to match altLabel
        if (locations.isEmpty()) {
            Literal altLabel = model.createLiteral(capitalized, "en");
            locations.addAll(model.listResourcesWithProperty(SKOS.altLabel, altLabel).toSet());
        }

        // try to match identifier
        if (locations.isEmpty())
            locations.addAll(model.listResourcesWithProperty(DC.identifier, location.toUpperCase()).toSet());

        // try to match authority code
        if (locations.isEmpty())
            locations.addAll(model.listResourcesWithProperty(AUTHORITY.authorityCode, location.toUpperCase()).toSet());

        // try to match legacy code (eg EN)
        if (locations.isEmpty())
            model.listResourcesWithProperty(AUTHORITY.legacyCode, location.toUpperCase()).toSet()
                .forEach(mappedCode -> locations.addAll(model.listResourcesWithProperty(AUTHORITY.opMappedCode, mappedCode).toSet()));


        return locations.isEmpty()
            ? Optional.empty()
            : Optional.of(locations.iterator().next());
    }

    /**
     * Converts a given {@link DateConstraint} into a Jena expression.
     *
     * @param dateConstraint The date range to convert
     * @param startDateVar   Start of the date range, included in the interval
     * @param endDateVar     End of the date range, not included in the interval
     * @return An Expression encoding the given DateConstraint
     */
    static Expr getDateRangeConstraint(DateConstraint dateConstraint, Var startDateVar, Var endDateVar) {

        Expr startDateConstraint = null;
        Expr endDateConstraint = null;

        LocalDateTime startDateTime = parseDate(dateConstraint.getStartDate(), true);
        if (startDateTime != null) {
            String startDateFormatted = startDateTime.format(DateTimeFormatter.ISO_DATE_TIME);
            startDateConstraint = new E_GreaterThanOrEqual(new ExprVar(startDateVar), new NodeValueDT(startDateFormatted, NodeFactory.createLiteral(startDateFormatted, XSDDatatype.XSDdateTime)));
        }

        LocalDateTime endDateTime = parseDate(dateConstraint.getEndDate(), false);
        if (endDateTime != null) {
            String endDateFormatted = endDateTime.format(DateTimeFormatter.ISO_DATE_TIME);
            endDateConstraint = new E_LessThan(new ExprVar(endDateVar), new NodeValueDT(endDateFormatted, NodeFactory.createLiteral(endDateFormatted, XSDDatatype.XSDdateTime)));
        }

        if (startDateConstraint != null && endDateConstraint == null)
            return startDateConstraint;

        if (startDateConstraint == null && endDateConstraint != null)
            return endDateConstraint;

        return new E_LogicalAnd(startDateConstraint, endDateConstraint);
    }

    /**
     * Attempts to parse a string containing some sort of date value into a LocalDate object for further processing.
     * FIXME this is only package private for testing, should be class private
     *
     * @param date        contains some date information, like 'today', '1960', 'last year' etc
     * @param isStartDate denotes if date is a start- (true) or end date (false). This is relevant for setting the correct month/day if only a year is given.
     *                    Example:  If datasets starting in 1960 are requested, the date must be set to the 1st of January.
     *                    Contrary, if datasets until 1960 are requested, the date must be set to 31st of December
     * @return a LocalDate containing the input date
     */
    static LocalDateTime parseDate(String date, boolean isStartDate) {

        final LocalTime startTime = LocalTime.of(0, 0, 0);
        final LocalTime endTime = LocalTime.of(23, 59, 59);

        if (date == null)
            return null;

        // TODO add more date patterns
        // ISO 8601 full date
        if (date.matches("\\d{4}-([0]\\d|1[0-2])-([0-2]\\d|3[01])")) {
            LocalDate localDate = LocalDate.parse(date);
            return isStartDate
                ? LocalDateTime.of(localDate, startTime)
                : LocalDateTime.of(localDate, endTime);
        }

        // TODO add more date patterns
        // ISO 8601 date w/o day
        if (date.matches("\\d{4}-([0]\\d|1[0-2])")) {
            YearMonth yearMonth = YearMonth.parse(date);
            return isStartDate
                ? LocalDateTime.of(yearMonth.getYear(), yearMonth.getMonth(), 1, 0, 0, 0)
                : LocalDateTime.of(yearMonth.getYear(), yearMonth.getMonth(), yearMonth.lengthOfMonth(), 23, 59, 59);
        }

        // TODO add more date patterns
        //  eg 1960s, last decade
        // ISO 8601 year only
        if (date.matches("\\d{4}")) {
            Year year = Year.parse(date);
            return isStartDate
                ? LocalDateTime.of(LocalDate.of(year.getValue(), 1, 1), startTime)    // 1st of January
                : LocalDateTime.of(LocalDate.of(year.getValue(), 12, 31), endTime);   // 31st of December
        }


        if ("today".equals(date.toLowerCase()))
            return isStartDate
                ? LocalDateTime.of(LocalDate.now(), startTime)
                : LocalDateTime.of(LocalDate.now(), endTime);

        if ("yesterday".equals(date.toLowerCase()))
            return isStartDate
                ? LocalDateTime.of(LocalDate.now().minusDays(1), startTime)
                : LocalDateTime.of(LocalDate.now().minusDays(1), endTime);


        final String regexPast = "(last|previous|past)";

        if (date.toLowerCase().matches(regexPast + " week"))
            return isStartDate
                ? LocalDateTime.of(LocalDate.now().minusWeeks(1), startTime)
                : LocalDateTime.of(LocalDate.now().minusWeeks(1), endTime);

        if (date.toLowerCase().matches(regexPast + " month")) {
            LocalDate localDate = LocalDate.now().minusMonths(1);
            return isStartDate
                ? LocalDateTime.of(localDate.withDayOfMonth(1), startTime)
                : LocalDateTime.of(localDate.withDayOfMonth(localDate.lengthOfMonth()), endTime);
        }

        if (date.toLowerCase().matches(regexPast + " year")) {
            LocalDate localDate = LocalDate.now().minusYears(1);
            return isStartDate
                ? LocalDateTime.of(localDate.withMonth(1).withDayOfMonth(1), startTime)      // 1st of January
                : LocalDateTime.of(localDate.withMonth(12).withDayOfMonth(31), endTime);   // 31st of December
        }


        if (date.toLowerCase().matches(regexPast + " (mon|monday)"))
            return getPreviousNamedDay(DayOfWeek.MONDAY, isStartDate);

        if (date.toLowerCase().matches(regexPast + " (tue|tuesday)"))
            return getPreviousNamedDay(DayOfWeek.TUESDAY, isStartDate);

        if (date.toLowerCase().matches(regexPast + " (wed|wednesday)"))
            return getPreviousNamedDay(DayOfWeek.WEDNESDAY, isStartDate);

        if (date.toLowerCase().matches(regexPast + " (thu|thursday)"))
            return getPreviousNamedDay(DayOfWeek.THURSDAY, isStartDate);

        if (date.toLowerCase().matches(regexPast + " (fri|friday)"))
            return getPreviousNamedDay(DayOfWeek.FRIDAY, isStartDate);

        if (date.toLowerCase().matches(regexPast + " (sat|saturday)"))
            return getPreviousNamedDay(DayOfWeek.SATURDAY, isStartDate);

        if (date.toLowerCase().matches(regexPast + " (sun|sunday)"))
            return getPreviousNamedDay(DayOfWeek.SUNDAY, isStartDate);


        if (date.toLowerCase().matches(regexPast + " (jan|january)"))
            return getPreviousNamedMonth(Month.JANUARY, isStartDate);

        if (date.toLowerCase().matches(regexPast + " (feb|february)"))
            return getPreviousNamedMonth(Month.FEBRUARY, isStartDate);

        if (date.toLowerCase().matches(regexPast + " (mar|march)"))
            return getPreviousNamedMonth(Month.MARCH, isStartDate);

        if (date.toLowerCase().matches(regexPast + " (apr|april)"))
            return getPreviousNamedMonth(Month.APRIL, isStartDate);

        if (date.toLowerCase().matches(regexPast + " may"))
            return getPreviousNamedMonth(Month.MAY, isStartDate);

        if (date.toLowerCase().matches(regexPast + " (jun|june)"))
            return getPreviousNamedMonth(Month.JUNE, isStartDate);

        if (date.toLowerCase().matches(regexPast + " (jul|july)"))
            return getPreviousNamedMonth(Month.JULY, isStartDate);

        if (date.toLowerCase().matches(regexPast + " (aug|august)"))
            return getPreviousNamedMonth(Month.AUGUST, isStartDate);

        if (date.toLowerCase().matches(regexPast + " (sep|september)"))
            return getPreviousNamedMonth(Month.SEPTEMBER, isStartDate);

        if (date.toLowerCase().matches(regexPast + " (oct|october)"))
            return getPreviousNamedMonth(Month.OCTOBER, isStartDate);

        if (date.toLowerCase().matches(regexPast + " (nov|november)"))
            return getPreviousNamedMonth(Month.NOVEMBER, isStartDate);

        if (date.toLowerCase().matches(regexPast + " (dec|december)"))
            return getPreviousNamedMonth(Month.DECEMBER, isStartDate);


        Matcher lastWeeks = Pattern.compile("(" + regexPast + " \\d+ week(s?)) | (\\d+ week(s?) ago)", Pattern.CASE_INSENSITIVE).matcher(date);

        if (lastWeeks.find())
            return isStartDate
                ? LocalDateTime.of(LocalDate.now().minusWeeks(Integer.parseInt(lastWeeks.group())), startTime)
                : LocalDateTime.of(LocalDate.now().minusWeeks(Integer.parseInt(lastWeeks.group())), endTime);


        Matcher lastMonths = Pattern.compile("(" + regexPast + " \\d+ month(s?)) | (\\d+ month(s?) ago)", Pattern.CASE_INSENSITIVE).matcher(date);

        if (lastMonths.find()) {
            LocalDate localDate = LocalDate.now().minusMonths(Integer.valueOf(lastMonths.group()));
            return isStartDate
                ? LocalDateTime.of(localDate, startTime)
                : LocalDateTime.of(localDate, endTime);
        }

        Matcher lastYears = Pattern.compile("(" + regexPast + " \\d+ year(s?)) | (\\d+ year(s?) ago)", Pattern.CASE_INSENSITIVE).matcher(date);

        if (lastYears.find()) {
            LocalDate localDate = LocalDate.now().minusYears(Integer.parseInt(lastYears.group()));
            return isStartDate
                ? LocalDateTime.of(localDate, startTime)          // 1st of January, 00:00:00
                : LocalDateTime.of(localDate, endTime);    // 31st of December, 23:59:59
        }

        return null;
    }

    /**
     * Helper method to return a LocalDateTime containing the previous named day, for example 'last monday'
     *
     * @param dayOfWeek the weekday of which the previous date is to be determined
     * @param isStartDate flag indicating if the time shall be an upper or lower bound, relevant for the time fields
     * @return a LocalDateTime object containing the previous named day
     */
    private static LocalDateTime getPreviousNamedDay(DayOfWeek dayOfWeek, boolean isStartDate) {
        return isStartDate
            ? LocalDateTime.of(LocalDate.now().with(TemporalAdjusters.previous(dayOfWeek)), LocalTime.of(0, 0, 0))
            : LocalDateTime.of(LocalDate.now().with(TemporalAdjusters.previous(dayOfWeek)), LocalTime.of(23, 59, 59));
    }

    /**
     * Helper method to return a LocalDateTime containing the previous named day, for example 'last monday'
     *
     * @param month the month for which the previous date is to be determined
     * @param isStartDate flg indicating if the time shall be an upper or lower bound, relevant for the day and time fields
     * @return a LocalDateTime object containing the previous named day
     */
    private static LocalDateTime getPreviousNamedMonth(Month month, boolean isStartDate) {
        LocalDate lastMonth = LocalDate.now().getMonthValue() > month.getValue()
            ? LocalDate.now().withMonth(month.getValue())
            : LocalDate.now().minusYears(1).withMonth(month.getValue());

        return isStartDate
            ? LocalDateTime.of(lastMonth.withDayOfMonth(1), LocalTime.of(0, 0, 0))
            : LocalDateTime.of(lastMonth.withDayOfMonth(lastMonth.lengthOfMonth()), LocalTime.of(23, 59, 59));
    }

    public static NodeValue bifContains(NodeValue string, NodeValue match) {
        NodeFunctions.checkTwoArgumentStringLiterals("bif:contains", string, match);
        String lex1 = string.asNode().getLiteralLexicalForm();
        String lex2 = match.asNode().getLiteralLexicalForm();
        boolean x = StrUtils.contains(lex1, lex2);
        return NodeValue.booleanReturn(x);
    }

}
