package de.fhg.fokus.edp.chatbot_sparql_service;

import de.fhg.fokus.edp.chatbot_sparql_service.model.Dataset;
import de.fhg.fokus.edp.chatbot_sparql_service.model.DatasetRequest;
import de.fhg.fokus.edp.chatbot_sparql_service.model.rdf.E_BifContains;
import de.fhg.fokus.edp.chatbot_sparql_service.model.rdf.SCHEMA;
import io.vertx.core.AbstractVerticle;
import io.vertx.core.Future;
import io.vertx.core.eventbus.Message;
import io.vertx.core.json.Json;
import io.vertx.core.json.JsonArray;
import io.vertx.core.json.JsonObject;
import io.vertx.ext.web.client.WebClient;
import io.vertx.ext.web.client.predicate.ResponsePredicate;
import org.apache.jena.arq.querybuilder.SelectBuilder;
import org.apache.jena.query.Query;
import org.apache.jena.query.QueryExecution;
import org.apache.jena.query.QueryExecutionFactory;
import org.apache.jena.query.QuerySolution;
import org.apache.jena.rdf.model.Model;
import org.apache.jena.rdf.model.ModelFactory;
import org.apache.jena.sparql.core.Var;
import org.apache.jena.sparql.expr.*;
import org.apache.jena.sparql.expr.nodevalue.NodeValueNode;
import org.apache.jena.sparql.expr.nodevalue.NodeValueString;
import org.apache.jena.vocabulary.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.stream.Collectors;

import static de.fhg.fokus.edp.chatbot_sparql_service.ApplicationConfig.*;
import static de.fhg.fokus.edp.chatbot_sparql_service.SparqlUtils.*;

public class SparqlVerticle extends AbstractVerticle {

    private static final Logger LOG = LoggerFactory.getLogger(SparqlVerticle.class);

    private String sparqlUrl;

    private static final String BIF_QUERY = "select * where {\n" +
        "  ?s a dcat:Dataset ;\n" +
        "  ?t dct:title .\n" +
        "  ?t bif:contains(\"\")\n" +
        "} LIMIT 1";

    private boolean supportsBif;

    private Model languageModel;
    private Model fileTypeModel;
    private Model spatialModel;

    @Override
    public void start(Future<Void> startFuture) {

        languageModel = ModelFactory.createDefaultModel();
        fileTypeModel = ModelFactory.createDefaultModel();
        spatialModel = ModelFactory.createDefaultModel();

        try {
            loadModels();

            sparqlUrl = config().getString(ENV_SPARQL_URL, DEFAULT_SPARQL_URL);

            // method future never fails
            getServerBifSupport().setHandler(bifSupportHandler -> {
                supportsBif = bifSupportHandler.result();

                vertx.eventBus().consumer(SPARQL_VERTICLE_ADDRESS, this::handleSearchRequest);

                startFuture.complete();
            });
        } catch (IOException e) {
            startFuture.fail(e);
        }
    }

    /**
     * Checks if Virtuoso's feature bif:contains can be used for string matching, which is expected to be faster.
     * The feature can be toggled manually by setting the @link() variable before launch.
     * If it is not set, a query including bif:contains is sent to the configured SPARQL endpoint.
     * Depending on the server's response, bif:contains is enabled or not.
     *
     * @return a Future containing the result of the check described.
     */
    private Future<Boolean> getServerBifSupport() {
        Future<Boolean> completionFuture = Future.future();

        if (config().getBoolean(ENV_ENABLE_BIF) != null) {
            completionFuture.complete(config().getBoolean(ENV_ENABLE_BIF));
        } else {
            WebClient.create(vertx).getAbs(sparqlUrl)
                .putHeader("Accept", "application/rdf+xml")
                .addQueryParam("query", BIF_QUERY)
                .expect(ResponsePredicate.SC_OK)
                .send(sendHandler ->
                    completionFuture.complete(sendHandler.succeeded()));
        }

        return completionFuture;
    }

    private void handleSearchRequest(Message<String> message) {
        DatasetRequest request = Json.decodeValue(message.body(), DatasetRequest.class);
        LOG.debug("Received {}", request);

        buildSparqlQuery(request).setHandler(queryHandler -> {
            if (queryHandler.succeeded()) {
                Query query = queryHandler.result();

                getRemoteDatasets(query).setHandler(resultHandler -> {
                    if (resultHandler.succeeded()) {
                        JsonArray result = new JsonArray(Json.encode(resultHandler.result()));
                        JsonObject reply = new JsonObject()
                            .put("success", true)
                            .put("result", result);
                        message.reply(reply);
                    } else {
                        JsonObject reply = new JsonObject()
                            .put("success", false)
                            .put("message", "Failed to complete request: " + resultHandler.cause());
                        message.reply(reply);
                    }
                });
            } else {
                JsonObject reply = new JsonObject()
                    .put("success", false)
                    .put("message", "Failed to execute search request: " + queryHandler.cause());
                message.reply(reply);
            }
        });
    }

    /**
     * Recursively fetches datasets from the configured SPARQL endpoint, repeating for as long as either
     * * the set number of resolvable datasets has been found or
     * * no more results are returned by the SPARQL endpoint
     *
     * @param query The prebuilt query. Limit and offset are added dynamically to this query
     * @return A List of resolvable datasets
     */
    private Future<List<Dataset>> getRemoteDatasets(Query query) {
        Future<List<Dataset>> completionFuture = Future.future();

        String sparqlUrl = config().getString(ENV_SPARQL_URL, DEFAULT_SPARQL_URL);

        LOG.info("Query: {}", query);

        // send SPARQL query to configured endpoint
        try (QueryExecution queryExecution = QueryExecutionFactory.sparqlService(sparqlUrl, query)) {

            List<QuerySolution> querySolutions = new ArrayList<>();
            queryExecution.execSelect().forEachRemaining(querySolutions::add);

            List<Dataset> datasets = new ArrayList<>();

            // retrieve list of datasets from result
            querySolutions.forEach(solution -> {

                // construct new dataset from query solution
                Dataset dataset = new Dataset();
                dataset.setTitle(solution.getLiteral("title").getString());

                // small hack as EDP only has certificates for the full URL
                String url = new StringBuilder(solution.getResource("dataset").getURI())
                    .insert(8, "www.")
                    .toString();

                dataset.setUrl(url);

                // check if dataset is resolvable, else discard
                datasets.add(dataset);
            });

            completionFuture.complete(datasets);

        } catch (Exception e) {
            LOG.warn("Failed to execute SPARQL request: {}", e.getMessage());
            completionFuture.complete(new ArrayList<>());
        }

        return completionFuture;
    }

    /**
     * Parses a given {@link DatasetRequest} and constructs a valid SPARQL query from its contents
     *
     * @param request The {@link DatasetRequest} to be parsed
     * @return A {@link Query} object honouring all constraints in the {@link DatasetRequest}
     */
    private Future<Query> buildSparqlQuery(DatasetRequest request) {
        Future<Query> queryFuture = Future.future();

        vertx.executeBlocking(queryBuilderHandler -> {

            try {
                Var dataset = Var.alloc("dataset");
                Var title = Var.alloc("title");
                Var description = Var.alloc("description");

                SelectBuilder selectBuilder = new SelectBuilder()
                    .addVar(dataset)
                    .addVar(title)
                    .addWhere(dataset, RDF.type, DCAT.Dataset)
                    .addWhere(dataset, DCTerms.title, title)
                    .addWhere(dataset, DCTerms.description, description)
                    .setDistinct(true);

                if (request.getDateConstraints() != null && !request.getDateConstraints().isEmpty()) {

                    Var periodOfTime = Var.alloc("periodOfTime");
                    Var startDate = Var.alloc("startDate");
                    Var endDate = Var.alloc("endDate");

                    List<Expr> dates = request.getDateConstraints().stream()
                        .filter(dateConstraint ->
                            !(dateConstraint.getStartDate() == null && dateConstraint.getEndDate() == null))
                        .map(dateConstraint ->
                            getDateRangeConstraint(dateConstraint, startDate, endDate)
                        ).collect(Collectors.toList());

                    if (!dates.isEmpty()) {
                        selectBuilder.addWhere(dataset, DCTerms.temporal, periodOfTime)
                            .addWhere(periodOfTime, SCHEMA.startDate, startDate)
                            .addWhere(periodOfTime, SCHEMA.endDate, endDate)
                            .addFilter(recursiveDisjunction(dates));
                    }
                }

                if (request.getEntities() != null) {
                    request.getEntities().forEach(entity -> {
                        if (entity.getValues() != null) {
                            // split value string by whitespaces
                            String[] values = entity.getValues().split("\\s+");

                            switch (entity.getType()) {
                                case SPATIAL:
                                    Var spatial = Var.alloc("spatial");
                                    List<Expr> locations = new ArrayList<>();

                                    for (String location : values) {
                                        getLocationByName(spatialModel, location)
                                            .ifPresent(loc -> locations.add(new E_Equals(new ExprVar(spatial), new NodeValueNode(loc.asNode()))));
                                    }

                                    if (!locations.isEmpty()) {
                                        selectBuilder.addWhere(dataset, DCTerms.spatial, spatial)
                                            .addFilter(!entity.getNegation()
                                                ? recursiveDisjunction(locations)
                                                : new E_LogicalNot(recursiveDisjunction(locations))
                                            );
                                    }

                                    break;

                                case FORMAT:
                                    Var distribution = Var.alloc("distribution");
                                    Var format = Var.alloc("format");
                                    Var mediaType = Var.alloc("mediaType");
                                    List<Expr> formats = new ArrayList<>();

                                    for (String fileType : values) {
                                        getMediaTypeByFormatName(fileTypeModel, fileType)
                                            .ifPresent(ft -> {
                                                formats.add(new E_Equals(new ExprVar(format), new NodeValueNode(ft.asNode())));
                                                formats.add(new E_Equals(new ExprVar(mediaType), new NodeValueNode(ft.asNode())));
                                            });
                                    }

                                    if (!formats.isEmpty()) {
                                        selectBuilder.addWhere(dataset, DCAT.distribution, distribution)
                                            .addWhere(distribution, DCTerms.format, format)
                                            .addWhere(distribution, DCAT.mediaType, mediaType)
                                            .addFilter(!entity.getNegation()
                                                ? recursiveDisjunction(formats)
                                                : new E_LogicalNot(recursiveDisjunction(formats))
                                            );
                                    }

                                    break;

                                case LANGUAGE:
                                    Var language = Var.alloc("language");
                                    List<Expr> languages = new ArrayList<>();

                                    for (String lang : values) {
                                        getLinguisticSystemByLanguageName(languageModel, lang)
                                            .ifPresent(l -> languages.add(new E_Equals(new ExprVar(language), new NodeValueNode(l.asNode()))));
                                    }

                                    if (!languages.isEmpty()) {
                                        // chain all specified languages using disjunction
                                        selectBuilder.addWhere(dataset, DCTerms.language, language)
                                            .addFilter(!entity.getNegation()
                                                ? recursiveDisjunction(languages)
                                                : new E_LogicalNot(recursiveDisjunction(languages))
                                            );
                                    }

                                    break;
                            }
                        }
                    });
                }

                if (request.getSearchTerms() != null && !request.getSearchTerms().isEmpty()) {
                    // all searchTerms shall either appear in a dataset's title or description
                    List<Expr> searchTerms = Arrays.stream(request.getSearchTerms().split(" "))
                        .map(searchTerm -> {
                            if (supportsBif) {
                                String bifConformVal = searchTerm.length() >= 4
                                    ? "'" + searchTerm + "*'"
                                    : "'" + searchTerm + "'";

                                E_BifContains titleContains = new E_BifContains(new ExprVar(title), new NodeValueString(bifConformVal));
                                E_BifContains descriptionContains = new E_BifContains(new ExprVar(description), new NodeValueString(bifConformVal));
                                return new E_LogicalOr(titleContains, descriptionContains);
                            } else {
                                E_StrContains titleContains = new E_StrContains(new ExprVar(title), new NodeValueString(searchTerm));
                                E_StrContains descriptionContains = new E_StrContains(new ExprVar(description), new NodeValueString(searchTerm));
                                return new E_LogicalOr(titleContains, descriptionContains);
                            }
                        }).collect(Collectors.toList());

                    // Virtuoso rejects chaining of bif:contains in filters. Therefore, SPARQL SubQueries are used.
                    // Error message: Virtuoso 37000 Error SP031: SPARQL compiler: No suitable triple pattern is found for a variable $description in special predicate bif:contains()
                    if (supportsBif) {
                        searchTerms.forEach(searchTerm ->
                            selectBuilder.addSubQuery(new SelectBuilder()
                                .addPrefix("bif", "bif:")
                                .setDistinct(true)

                                .addVar(dataset)
                                .addVar(title)
                                .addWhere(dataset, RDF.type, DCAT.Dataset)
                                .addWhere(dataset, DCTerms.title, title)
                                .addWhere(dataset, DCTerms.description, description))

                                .addFilter(searchTerm));
                    } else {
                        // all searchTerms need to appear somewhere, therefore conjunction is used
                        selectBuilder.addFilter(recursiveConjunction(searchTerms));
                    }

                    // remove duplicates stemming from multi language 'title' and 'description'
                    selectBuilder.addFilter(getLanguageFilter(title))
                        .addFilter(getLanguageFilter(description));
                }
                
                queryBuilderHandler.complete(selectBuilder.build());

            } catch (Exception e) {
                e.printStackTrace();
                LOG.error("Failed to build intent subquery: {}", e.getMessage());
                queryBuilderHandler.fail(e);
            }

        }, resultHandler -> {
            if (resultHandler.succeeded()) {
                queryFuture.complete((Query) resultHandler.result());
            } else {
                queryFuture.fail("Failed to build query for " + request.toString());
            }
        });

        return queryFuture;
    }

    private void loadModels() throws IOException {

        // load language model
        try (InputStream is = getClass().getClassLoader().getResourceAsStream("vocabularies/languages.rdf")) {
            if (is != null) {
                languageModel.read(is, null, "RDF/XML");
            } else {
                throw new IOException("Language file InputStream is null");
            }
        }

        // load file type model
        try (InputStream is = getClass().getClassLoader().getResourceAsStream("vocabularies/fileTypes.rdf")) {
            if (is != null) {
                fileTypeModel.read(is, null, "RDF/XML");
            } else {
                throw new IOException("File type InputStream is null");
            }
        }

        try (InputStream is = getClass().getClassLoader().getResourceAsStream("vocabularies/continents.rdf")) {
            if (is != null) {
                spatialModel.read(is, null, "RDF/XML");
            } else {
                throw new IOException("Continent file InputStream is null");
            }
        }

        try (InputStream is = getClass().getClassLoader().getResourceAsStream("vocabularies/countries.rdf")) {
            if (is != null) {
                spatialModel.read(is, null, "RDF/XML");
            } else {
                throw new IOException("Country file InputStream is null");
            }
        }

        try (InputStream is = getClass().getClassLoader().getResourceAsStream("vocabularies/places.rdf")) {
            if (is != null) {
                spatialModel.read(is, null, "RDF/XML");
            } else {
                throw new IOException("Place file InputStream is null");
            }
        }
    }
}
