package de.fhg.fokus.edp.chatbot_sparql_service;

import org.junit.jupiter.api.Test;

import java.time.*;
import java.time.temporal.TemporalAdjusters;

import static de.fhg.fokus.edp.chatbot_sparql_service.SparqlUtils.parseDate;
import static org.junit.jupiter.api.Assertions.assertNull;
import static org.junit.jupiter.api.Assertions.assertTrue;

class DateParserTest {

    @Test
    void testIsoYearAsStartDate() {
        assertTrue(LocalDateTime.of(2019, 1, 1, 0, 0, 0).isEqual(parseDate("2019", true)));
    }

    @Test
    void testIsoYearAsEndDate() {
        assertTrue(LocalDateTime.of(2019, 12, 31, 23, 59, 59).isEqual(parseDate("2019", false)));
    }

    @Test
    void testIsoYearMonthAsStartDate() {
        assertTrue(LocalDateTime.of(2019, 5, 1, 0, 0, 0).isEqual(parseDate("2019-05", true)));
    }

    @Test
    void testIsoYearMonthAsEndDate() {
        assertTrue(LocalDateTime.of(2019, 5, 31, 23, 59, 59).isEqual(parseDate("2019-05", false)));
    }

    @Test
    void testIsoFullDateAsStartDate() {
        assertTrue(LocalDateTime.of(2019, 5, 1, 0, 0, 0).isEqual(parseDate("2019-05-01", true)));
    }

    @Test
    void testIsoFullDateAsEndDate() {
        assertTrue(LocalDateTime.of(2019, 5, 1, 23, 59, 59).isEqual(parseDate("2019-05-01", false)));
    }

    @Test
    void testTodayAsStartDate() {
        assertTrue(LocalDateTime.of(LocalDate.now(), LocalTime.of(0, 0, 0)).isEqual(parseDate("today", true)));
    }

    @Test
    void testTodayAsEndDate() {
        assertTrue(LocalDateTime.of(LocalDate.now(), LocalTime.of(23, 59, 59)).isEqual(parseDate("today", false)));
    }

    @Test
    void testYesterdayAsStartDate() {
        assertTrue(LocalDateTime.of(LocalDate.now().minusDays(1), LocalTime.of(0, 0, 0)).isEqual(parseDate("yesterday", true)));
    }

    @Test
    void testYesterdayAsEndDate() {
        assertTrue(LocalDateTime.of(LocalDate.now().minusDays(1), LocalTime.of(23, 59, 59)).isEqual(parseDate("yesterday", false)));
    }

    @Test
    void testLastYearAsStartDate() {
        assertTrue(LocalDateTime.of(LocalDate.now().minusYears(1).withMonth(1).withDayOfMonth(1), LocalTime.of(0, 0, 0)).isEqual(parseDate("last year", true)));
    }

    @Test
    void testLastYearAsEndDate() {
        assertTrue(LocalDateTime.of(LocalDate.now().minusYears(1).withMonth(12).withDayOfMonth(31), LocalTime.of(23, 59, 59)).isEqual(parseDate("last year", false)));
    }

    @Test
    void testLastMonthAsStartDate() {
        assertTrue(LocalDateTime.of(LocalDate.now().minusMonths(1).withDayOfMonth(1), LocalTime.of(0, 0, 0)).isEqual(parseDate("last month", true)));
    }

    @Test
    void testLastMonthAsEndDate() {
        LocalDate lastMonth = LocalDate.now().minusMonths(1);
        assertTrue(LocalDateTime.of(lastMonth.withDayOfMonth(lastMonth.lengthOfMonth()), LocalTime.of(23, 59, 59)).isEqual(parseDate("last month", false)));
    }

    @Test
    void testPreviousNamedMonthAsStartDate() {
        LocalDate expectedMonth = LocalDate.now().getMonthValue() <= 5
            ? LocalDate.now().minusYears(1).withMonth(Month.MAY.getValue())
            : LocalDate.now().withMonth(Month.MAY.getValue());

        assertTrue(LocalDateTime.of(expectedMonth.withDayOfMonth(1), LocalTime.of(0, 0, 0)).isEqual(parseDate("last may", true)));
    }

    @Test
    void testPreviousNamedMonthAsEndDate() {
        LocalDate expectedMonth = LocalDate.now().getMonthValue() <= 5
            ? LocalDate.now().minusYears(1).withMonth(Month.MAY.getValue())
            : LocalDate.now().withMonth(Month.MAY.getValue());

        assertTrue(LocalDateTime.of(expectedMonth.withDayOfMonth(31), LocalTime.of(23, 59, 59)).isEqual(parseDate("last may", false)));
    }

    @Test
    void testPreviousNamedDayAsStartDate() {
        assertTrue(LocalDateTime.of(LocalDate.now().with(TemporalAdjusters.previous(DayOfWeek.WEDNESDAY)), LocalTime.of(0, 0, 0)).isEqual(parseDate("last wednesday", true)));
    }

    @Test
    void testPreviousNamedDayAsEndDate() {
        assertTrue(LocalDateTime.of(LocalDate.now().with(TemporalAdjusters.previous(DayOfWeek.WEDNESDAY)), LocalTime.of(23, 59, 59)).isEqual(parseDate("last wednesday", false)));
    }

    @Test
    void testLastWeekAsStartDate() {
        assertTrue(LocalDateTime.of(LocalDate.now().minusWeeks(1), LocalTime.of(0, 0, 0)).isEqual(parseDate("last week", true)));
    }

    @Test
    void testLastWeekAsEndDate() {
        assertTrue(LocalDateTime.of(LocalDate.now().minusWeeks(1), LocalTime.of(23, 59, 59)).isEqual(parseDate("last week", false)));
    }

    @Test
    void testInvalidDate() {
        assertNull(parseDate("abc", true));
    }
}
