package de.fhg.fokus.edp.chatbot_sparql_service;

import org.apache.jena.rdf.model.Model;
import org.apache.jena.rdf.model.ModelFactory;
import org.apache.jena.rdf.model.Resource;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.io.IOException;
import java.io.InputStream;
import java.util.Optional;

import static org.junit.jupiter.api.Assertions.*;

class SpatialParserTest {

    private static final String expectedContinent = "http://publications.europa.eu/resource/authority/continent/EUROPE";
    private static final String expectedCountry = "http://publications.europa.eu/resource/authority/country/GBR";
    private static final String expectedPlace = "http://publications.europa.eu/resource/authority/place/DEU_BER";

    private Model locationModel;

    @BeforeEach
    void setUp() {
        if (locationModel == null) {
            locationModel = ModelFactory.createDefaultModel();

            try (InputStream is = getClass().getClassLoader().getResourceAsStream("vocabularies/continents.rdf")) {
                if (is != null) {
                    locationModel.read(is, null, "RDF/XML");
                } else {
                    throw new IOException("Continent file InputStream is null");
                }
            } catch (IOException e) {
                fail(e);
            }

            try (InputStream is = getClass().getClassLoader().getResourceAsStream("vocabularies/countries.rdf")) {
                if (is != null) {
                    locationModel.read(is, null, "RDF/XML");
                } else {
                    throw new IOException("Country file InputStream is null");
                }
            } catch (IOException e) {
                fail(e);
            }

            try (InputStream is = getClass().getClassLoader().getResourceAsStream("vocabularies/places.rdf")) {
                if (is != null) {
                    locationModel.read(is, null, "RDF/XML");
                } else {
                    throw new IOException("Place file InputStream is null");
                }
            } catch (IOException e) {
                fail(e);
            }
        }
    }

    @Test
    void testContinent() {
        Optional<Resource> language = SparqlUtils.getLocationByName(locationModel, "europe");

        assertTrue(language.isPresent());
        assertEquals(expectedContinent, language.get().getURI());
    }

    @Test
    void testInvalidContinent() {
        assertFalse(SparqlUtils.getLocationByName(locationModel, "invalid_continent").isPresent());
    }

    @Test
    void testCountryPrefLabel() {
        checkCountry("United Kingdom");
    }

    @Test
    void testCountryLegacyCode() {
        checkCountry("UK");
    }

    @Test
    void testInvalidCountry() {
        assertFalse(SparqlUtils.getLocationByName(locationModel, "invalid_country").isPresent());
    }

    @Test
    void testPlacePrefLabel() {
        checkPlace("Berlin");
    }

    @Test
    void testPlaceAuthorityCode() {
        checkPlace("DEU_BER");
    }

    @Test
    void testInvalidPlace() {
        assertFalse(SparqlUtils.getLocationByName(locationModel, "invalid_place").isPresent());
    }

    private void checkCountry(String testCountry) {
        Optional<Resource> language = SparqlUtils.getLocationByName(locationModel, testCountry);

        assertTrue(language.isPresent());
        assertEquals(expectedCountry, language.get().getURI());
    }

    private void checkPlace(String testPlace) {
        Optional<Resource> language = SparqlUtils.getLocationByName(locationModel, testPlace);

        assertTrue(language.isPresent());
        assertEquals(expectedPlace, language.get().getURI());
    }
}
