package de.fhg.fokus.edp.chatbot_sparql_service.bot;

import de.fhg.fokus.edp.chatbot_sparql_service.ApplicationConfig;
import de.fhg.fokus.edp.chatbot_sparql_service.model.EntityType;
import io.vertx.core.Future;
import io.vertx.core.Vertx;
import io.vertx.core.buffer.Buffer;
import io.vertx.core.http.HttpServerOptions;
import io.vertx.core.json.JsonArray;
import io.vertx.core.json.JsonObject;
import io.vertx.core.net.NetClient;
import io.vertx.ext.web.Router;
import io.vertx.ext.web.handler.BodyHandler;
import io.vertx.junit5.Timeout;
import io.vertx.junit5.VertxExtension;
import io.vertx.junit5.VertxTestContext;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.TestInstance;
import org.junit.jupiter.api.extension.ExtendWith;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

@ExtendWith(VertxExtension.class)
@TestInstance(TestInstance.Lifecycle.PER_CLASS)
@AssumeEnvVariable(envVariables = {ApplicationConfig.ENV_CHATSCRIPT_HOST, ApplicationConfig.ENV_CHATSCRIPT_PORT})
@Timeout(value = 2, timeUnit = TimeUnit.MINUTES)
class ChatScriptTest {

    private static final Logger LOG = LoggerFactory.getLogger(ChatScriptTest.class);

    private static final String KEY_SEARCH_TERMS = "searchTerms";
    private static final String KEY_ENTITIES = "entities";
    private static final String KEY_DATE_CONSTRAINTS = "dateConstraints";

    private static final String SEARCH_TERM = "traffic";
    private static final List<String> SEARCH_TERM_INDICATOR = Arrays.asList("", "I am interested in data about", "Show me datasets with topic");

    private static final List<String> NEGATION = Arrays.asList("not", "but not", "without", "except", "exclude");

    private static final String FORMAT = "xls";
    private static final List<String> FORMAT_INDICATOR = Arrays.asList("as", "of format", "with type");

    private static final String SPATIAL = "Germany";
    private static final List<String> SPATIAL_INDICATOR = Arrays.asList("from", "regarding", "with spatial");

    private static final String LANGUAGE = "English";
    private static final List<String> LANGUAGE_INDICATOR = Arrays.asList("in", "of language", "with language");

    // TODO test all values
    private static final List<String> TEMPORAL_VALUES = Arrays.asList("today", "yesterday", "last monday", "last tuesday", "last wednesday", "last thursday", "last friday", "last saturday", "last sunday", "last week", "last month", "last year");
    private static final List<String> START_DATE_INDICATORS = Arrays.asList("since", "with start date");
    private static final List<String> END_DATE_INDICATORS = Arrays.asList("till", "until", "to", "with end date");

    // default host/port cannot be set in NetClientOptions
    private NetClient netClient;
    private String chatScriptHost;
    private int chatScriptPort;

    private TestCase testCase;
    private Future<Void> testCompleted;

    @BeforeAll
    void setUp(Vertx vertx, VertxTestContext testContext) {
        int applicationPort;

        try {
            applicationPort = Integer.parseInt(System.getenv(ApplicationConfig.ENV_APPLICATION_PORT));
            LOG.info("Testing with server port [{}]", applicationPort);
        } catch (NumberFormatException e) {
            applicationPort = ApplicationConfig.DEFAULT_APPLICATION_PORT;
            LOG.info("No port configured, using default port [{}]", applicationPort);
        }

        // Assumed variables containing 'port' are automatically checked for validity
        chatScriptHost = System.getenv(ApplicationConfig.ENV_CHATSCRIPT_HOST);
        chatScriptPort = Integer.parseInt(System.getenv(ApplicationConfig.ENV_CHATSCRIPT_PORT));

        Router router = Router.router(vertx);
        router.route("/search").handler(BodyHandler.create()).handler(routingContext -> {
            /*
             * Asserts cannot be used for checking test conditions, as VertxTestContext throws an uncatchable exception on assertion failure.
             * This prevents completing the test future which causes all following tests to run into timeouts.
             *
             * JsonObjects are used for comparison, since their 'equals' method deep inspects all values for equality instead of simply using the objects reference
             */

            if (routingContext.getBody() != null) {
                LOG.debug("Received request: [{}]", routingContext.getBodyAsString());

                // direct comparison would fail as ChatScript dataset request contains keys irrelevant for testing
                JsonObject sanitizedChatScriptResponse = new JsonObject();
                sanitizedChatScriptResponse.put("searchTerms", routingContext.getBodyAsJson().getString("searchTerms"));
                sanitizedChatScriptResponse.put("entities", routingContext.getBodyAsJson().getJsonArray("entities"));
                sanitizedChatScriptResponse.put("dateConstraints", routingContext.getBodyAsJson().getJsonArray("dateConstraints"));

                if (testCase.getExpectedResult().equals(sanitizedChatScriptResponse)) {
                    testCompleted.complete();
                } else {
                    testCompleted.fail("ChatScript response not equal to expected result");
                    LOG.info("Utterance: {}", testCase.getUtterance());
                    LOG.info("Expected: {}", testCase.getExpectedResult().encode());
                    LOG.info("Actual: {}", sanitizedChatScriptResponse.encode());
                }
            } else {
                testCompleted.fail("ChatScript response contained no body");
            }

            // always close connection
            routingContext.response().setStatusCode(200).end();
        });

        vertx.createHttpServer(new HttpServerOptions().setPort(applicationPort))
            .requestHandler(router).listen(testContext.succeeding());

        netClient = vertx.createNetClient();
        testContext.completeNow();
    }

    @Test
    void testSearchTerm(VertxTestContext testContext) {
        List<TestCase> testCases = getIndicatorValueCombinations(SEARCH_TERM_INDICATOR, SEARCH_TERM).stream().map(utterance -> {
            JsonObject expectedResponse = new JsonObject()
                .put(KEY_SEARCH_TERMS, SEARCH_TERM)
                .put(KEY_ENTITIES, new JsonArray())
                .put(KEY_DATE_CONSTRAINTS, new JsonArray());

            return new TestCase(utterance, expectedResponse);
        }).collect(Collectors.toList());

        LOG.info("Testing search terms with [{}] utterances", testCases.size());
        handleTestResult(testRecursively(testCases), testContext);
    }

    @Test
    void testFormat(VertxTestContext testContext) {
        List<String> searchTermCombinations = getIndicatorValueCombinations(SEARCH_TERM_INDICATOR, SEARCH_TERM);
        List<String> formatCombinations = getIndicatorValueCombinations(FORMAT_INDICATOR, FORMAT);

        List<TestCase> testCases = getCartesianProduct(searchTermCombinations, formatCombinations).stream().map(utterance -> {
            JsonObject expectedResponse = new JsonObject()
                .put(KEY_SEARCH_TERMS, SEARCH_TERM)
                .put(KEY_ENTITIES, new JsonArray().add(buildEntity(EntityType.FORMAT, FORMAT, false)))
                .put(KEY_DATE_CONSTRAINTS, new JsonArray());

            return new TestCase(utterance, expectedResponse);
        }).collect(Collectors.toList());

        LOG.info("Testing format with [{}] utterances", testCases.size());
        handleTestResult(testRecursively(testCases), testContext);
    }

    @Test
    void testFormatNegated(VertxTestContext testContext) {
        List<String> searchTermCombinations = getIndicatorValueCombinations(SEARCH_TERM_INDICATOR, SEARCH_TERM);
        List<String> negatedFormatIndicator = getCartesianProduct(NEGATION, FORMAT_INDICATOR);
        List<String> formatCombinations = getIndicatorValueCombinations(negatedFormatIndicator, FORMAT);

        List<TestCase> testCases = getCartesianProduct(searchTermCombinations, formatCombinations).stream().map(utterance -> {
            JsonObject expectedResponse = new JsonObject()
                .put(KEY_SEARCH_TERMS, SEARCH_TERM)
                .put(KEY_ENTITIES, new JsonArray().add(buildEntity(EntityType.FORMAT, FORMAT, true)))
                .put(KEY_DATE_CONSTRAINTS, new JsonArray());

            return new TestCase(utterance, expectedResponse);
        }).collect(Collectors.toList());

        LOG.info("Testing format negated with [{}] utterances", testCases.size());
        handleTestResult(testRecursively(testCases), testContext);
    }

    @Test
    void testSpatial(VertxTestContext testContext) {
        List<String> searchTermCombinations = getIndicatorValueCombinations(SEARCH_TERM_INDICATOR, SEARCH_TERM);
        List<String> spatialCombinations = getIndicatorValueCombinations(SPATIAL_INDICATOR, SPATIAL);

        List<TestCase> testCases = getCartesianProduct(searchTermCombinations, spatialCombinations).stream().map(utterance -> {
            JsonObject expectedResponse = new JsonObject()
                .put(KEY_SEARCH_TERMS, SEARCH_TERM)
                .put(KEY_ENTITIES, new JsonArray().add(buildEntity(EntityType.SPATIAL, SPATIAL, false)))
                .put(KEY_DATE_CONSTRAINTS, new JsonArray());

            return new TestCase(utterance, expectedResponse);
        }).collect(Collectors.toList());

        handleTestResult(testRecursively(testCases), testContext);
    }

    @Test
    void testSpatialNegated(VertxTestContext testContext) {
        List<String> searchTermCombinations = getIndicatorValueCombinations(SEARCH_TERM_INDICATOR, SEARCH_TERM);
        List<String> negatedSpatialIndicator = getCartesianProduct(NEGATION, SPATIAL_INDICATOR);
        List<String> spatialCombinations = getIndicatorValueCombinations(negatedSpatialIndicator, SPATIAL);

        List<TestCase> testCases = getCartesianProduct(searchTermCombinations, spatialCombinations).stream().map(utterance -> {
            JsonObject expectedResponse = new JsonObject()
                .put(KEY_SEARCH_TERMS, SEARCH_TERM)
                .put(KEY_ENTITIES, new JsonArray().add(buildEntity(EntityType.SPATIAL, SPATIAL, true)))
                .put(KEY_DATE_CONSTRAINTS, new JsonArray());

            return new TestCase(utterance, expectedResponse);
        }).collect(Collectors.toList());

        LOG.info("Testing spatial negated with [{}] utterances", testCases.size());
        handleTestResult(testRecursively(testCases), testContext);
    }

    @Test
    void testLanguage(VertxTestContext testContext) {
        List<String> searchTermCombinations = getIndicatorValueCombinations(SEARCH_TERM_INDICATOR, SEARCH_TERM);
        List<String> languageCombinations = getIndicatorValueCombinations(LANGUAGE_INDICATOR, LANGUAGE);

        List<TestCase> testCases = getCartesianProduct(searchTermCombinations, languageCombinations).stream().map(utterance -> {
            JsonObject expectedResponse = new JsonObject()
                .put(KEY_SEARCH_TERMS, SEARCH_TERM)
                .put(KEY_ENTITIES, new JsonArray().add(buildEntity(EntityType.LANGUAGE, LANGUAGE, false)))
                .put(KEY_DATE_CONSTRAINTS, new JsonArray());

            return new TestCase(utterance, expectedResponse);
        }).collect(Collectors.toList());

        LOG.info("Testing language with [{}] utterances", testCases.size());
        handleTestResult(testRecursively(testCases), testContext);
    }

    @Test
    void testLanguageNegated(VertxTestContext testContext) {
        List<String> searchTermCombinations = getIndicatorValueCombinations(SEARCH_TERM_INDICATOR, SEARCH_TERM);
        List<String> negatedLanguageIndicator = getCartesianProduct(NEGATION, LANGUAGE_INDICATOR);
        List<String> languageCombinations = getIndicatorValueCombinations(negatedLanguageIndicator, LANGUAGE);

        List<TestCase> testCases = getCartesianProduct(searchTermCombinations, languageCombinations).stream().map(utterance -> {
            JsonObject expectedResponse = new JsonObject()
                .put(KEY_SEARCH_TERMS, SEARCH_TERM)
                .put(KEY_ENTITIES, new JsonArray().add(buildEntity(EntityType.LANGUAGE, LANGUAGE, true)))
                .put(KEY_DATE_CONSTRAINTS, new JsonArray());

            return new TestCase(utterance, expectedResponse);
        }).collect(Collectors.toList());

        LOG.info("Testing language negated with [{}] utterances", testCases.size());
        handleTestResult(testRecursively(testCases), testContext);
    }

    @Test
    void testFormatAndSpatial(VertxTestContext testContext) {
        List<String> searchTermCombinations = getIndicatorValueCombinations(SEARCH_TERM_INDICATOR, SEARCH_TERM);

        List<String> formatCombinations = getIndicatorValueCombinations(FORMAT_INDICATOR, FORMAT);
        List<String> spatialCombinations = getIndicatorValueCombinations(SPATIAL_INDICATOR, SPATIAL);
        List<String> entityCombinations = getAllCombinations(formatCombinations, spatialCombinations);

        List<TestCase> testCases = getCartesianProduct(searchTermCombinations, entityCombinations).stream().map(utterance -> {
            JsonObject expectedResponse = new JsonObject()
                .put(KEY_SEARCH_TERMS, SEARCH_TERM)
                .put(KEY_ENTITIES, new JsonArray()
                    .add(buildEntity(EntityType.FORMAT, FORMAT, false))
                    .add(buildEntity(EntityType.SPATIAL, SPATIAL, false)))
                .put(KEY_DATE_CONSTRAINTS, new JsonArray());

            return new TestCase(utterance, expectedResponse);
        }).collect(Collectors.toList());

        LOG.info("Testing format and spatial with [{}] utterances", testCases.size());
        handleTestResult(testRecursively(testCases), testContext);
    }

    @Test
    void testFormatNegatedAndSpatial(VertxTestContext testContext) {
        List<String> searchTermCombinations = getIndicatorValueCombinations(SEARCH_TERM_INDICATOR, SEARCH_TERM);

        List<String> negatedFormatCombinations = getCartesianProduct(NEGATION, getIndicatorValueCombinations(FORMAT_INDICATOR, FORMAT));
        List<String> spatialCombinations = getIndicatorValueCombinations(SPATIAL_INDICATOR, SPATIAL);
        List<String> entityCombinations = getAllCombinations(negatedFormatCombinations, spatialCombinations);

        List<TestCase> testCases = getCartesianProduct(searchTermCombinations, entityCombinations).stream().map(utterance -> {
            JsonObject expectedResponse = new JsonObject()
                .put(KEY_SEARCH_TERMS, SEARCH_TERM)
                .put(KEY_ENTITIES, new JsonArray()
                    .add(buildEntity(EntityType.FORMAT, FORMAT, true))
                    .add(buildEntity(EntityType.SPATIAL, SPATIAL, false)))
                .put(KEY_DATE_CONSTRAINTS, new JsonArray());

            return new TestCase(utterance, expectedResponse);
        }).collect(Collectors.toList());

        LOG.info("Testing format negated and spatial with [{}] utterances", testCases.size());
        handleTestResult(testRecursively(testCases), testContext);
    }

    @Test
    void testFormatAndSpatialNegated(VertxTestContext testContext) {
        List<String> searchTermCombinations = getIndicatorValueCombinations(SEARCH_TERM_INDICATOR, SEARCH_TERM);

        List<String> formatCombinations = getIndicatorValueCombinations(FORMAT_INDICATOR, FORMAT);
        List<String> negatedSpatialCombinations = getCartesianProduct(NEGATION, getIndicatorValueCombinations(SPATIAL_INDICATOR, SPATIAL));
        List<String> entityCombinations = getAllCombinations(formatCombinations, negatedSpatialCombinations);

        List<TestCase> testCases = getCartesianProduct(searchTermCombinations, entityCombinations).stream().map(utterance -> {
            JsonObject expectedResponse = new JsonObject()
                .put(KEY_SEARCH_TERMS, SEARCH_TERM)
                .put(KEY_ENTITIES, new JsonArray()
                    .add(buildEntity(EntityType.FORMAT, FORMAT, false))
                    .add(buildEntity(EntityType.SPATIAL, SPATIAL, true)))
                .put(KEY_DATE_CONSTRAINTS, new JsonArray());

            return new TestCase(utterance, expectedResponse);
        }).collect(Collectors.toList());

        LOG.info("Testing format and spatial negated with [{}] utterances", testCases.size());
        handleTestResult(testRecursively(testCases), testContext);
    }

    @Test
    void testFormatNegatedAndSpatialNegated(VertxTestContext testContext) {
        List<String> searchTermCombinations = getIndicatorValueCombinations(SEARCH_TERM_INDICATOR, SEARCH_TERM);

        List<String> negatedFormatCombinations = getCartesianProduct(NEGATION, getIndicatorValueCombinations(FORMAT_INDICATOR, FORMAT));
        List<String> negatedSpatialCombinations = getCartesianProduct(NEGATION, getIndicatorValueCombinations(SPATIAL_INDICATOR, SPATIAL));
        List<String> entityCombinations = getAllCombinations(negatedFormatCombinations, negatedSpatialCombinations);

        List<TestCase> testCases = getCartesianProduct(searchTermCombinations, entityCombinations).stream().map(utterance -> {
            JsonObject expectedResponse = new JsonObject()
                .put(KEY_SEARCH_TERMS, SEARCH_TERM)
                .put(KEY_ENTITIES, new JsonArray()
                    .add(buildEntity(EntityType.FORMAT, FORMAT, true))
                    .add(buildEntity(EntityType.SPATIAL, SPATIAL, true)))
                .put(KEY_DATE_CONSTRAINTS, new JsonArray());

            return new TestCase(utterance, expectedResponse);
        }).collect(Collectors.toList());

        LOG.info("Testing format negated and spatial negated with [{}] utterances", testCases.size());
        handleTestResult(testRecursively(testCases), testContext);
    }

    @Test
    void testFormatAndLanguage(VertxTestContext testContext) {
        List<String> searchTermCombinations = getIndicatorValueCombinations(SEARCH_TERM_INDICATOR, SEARCH_TERM);

        List<String> formatCombinations = getIndicatorValueCombinations(FORMAT_INDICATOR, FORMAT);
        List<String> languageCombinations = getIndicatorValueCombinations(LANGUAGE_INDICATOR, LANGUAGE);
        List<String> entityCombinations = getAllCombinations(formatCombinations, languageCombinations);

        List<TestCase> testCases = getCartesianProduct(searchTermCombinations, entityCombinations).stream().map(utterance -> {
            JsonObject expectedResponse = new JsonObject()
                .put(KEY_SEARCH_TERMS, SEARCH_TERM)
                .put(KEY_ENTITIES, new JsonArray()
                    .add(buildEntity(EntityType.FORMAT, FORMAT, false))
                    .add(buildEntity(EntityType.LANGUAGE, LANGUAGE, false)))
                .put(KEY_DATE_CONSTRAINTS, new JsonArray());

            return new TestCase(utterance, expectedResponse);
        }).collect(Collectors.toList());

        LOG.info("Testing format and language with [{}] utterances", testCases.size());
        handleTestResult(testRecursively(testCases), testContext);
    }

    @Test
    void testFormatNegatedAndLanguage(VertxTestContext testContext) {
        List<String> searchTermCombinations = getIndicatorValueCombinations(SEARCH_TERM_INDICATOR, SEARCH_TERM);

        List<String> negatedFormatCombinations = getCartesianProduct(NEGATION, getIndicatorValueCombinations(FORMAT_INDICATOR, FORMAT));
        List<String> languageCombinations = getIndicatorValueCombinations(LANGUAGE_INDICATOR, LANGUAGE);
        List<String> entityCombinations = getAllCombinations(negatedFormatCombinations, languageCombinations);

        List<TestCase> testCases = getCartesianProduct(searchTermCombinations, entityCombinations).stream().map(utterance -> {
            JsonObject expectedResponse = new JsonObject()
                .put(KEY_SEARCH_TERMS, SEARCH_TERM)
                .put(KEY_ENTITIES, new JsonArray()
                    .add(buildEntity(EntityType.FORMAT, FORMAT, true))
                    .add(buildEntity(EntityType.LANGUAGE, LANGUAGE, false)))
                .put(KEY_DATE_CONSTRAINTS, new JsonArray());

            return new TestCase(utterance, expectedResponse);
        }).collect(Collectors.toList());

        LOG.info("Testing format negated and language with [{}] utterances", testCases.size());
        handleTestResult(testRecursively(testCases), testContext);
    }

    @Test
    void testFormatAndLanguageNegated(VertxTestContext testContext) {
        List<String> searchTermCombinations = getIndicatorValueCombinations(SEARCH_TERM_INDICATOR, SEARCH_TERM);

        List<String> formatCombinations = getIndicatorValueCombinations(FORMAT_INDICATOR, FORMAT);
        List<String> negatedLanguageCombinations = getCartesianProduct(NEGATION, getIndicatorValueCombinations(LANGUAGE_INDICATOR, LANGUAGE));
        List<String> entityCombinations = getAllCombinations(formatCombinations, negatedLanguageCombinations);

        List<TestCase> testCases = getCartesianProduct(searchTermCombinations, entityCombinations).stream().map(utterance -> {
            JsonObject expectedResponse = new JsonObject()
                .put(KEY_SEARCH_TERMS, SEARCH_TERM)
                .put(KEY_ENTITIES, new JsonArray()
                    .add(buildEntity(EntityType.FORMAT, FORMAT, false))
                    .add(buildEntity(EntityType.LANGUAGE, LANGUAGE, true)))
                .put(KEY_DATE_CONSTRAINTS, new JsonArray());

            return new TestCase(utterance, expectedResponse);
        }).collect(Collectors.toList());

        LOG.info("Testing format and language negated with [{}] utterances", testCases.size());
        handleTestResult(testRecursively(testCases), testContext);
    }

    @Test
    void testFormatNegatedAndLanguageNegated(VertxTestContext testContext) {
        List<String> searchTermCombinations = getIndicatorValueCombinations(SEARCH_TERM_INDICATOR, SEARCH_TERM);

        List<String> negatedFormatCombinations = getCartesianProduct(NEGATION, getIndicatorValueCombinations(FORMAT_INDICATOR, FORMAT));
        List<String> negatedLanguageCombinations = getCartesianProduct(NEGATION, getIndicatorValueCombinations(LANGUAGE_INDICATOR, LANGUAGE));
        List<String> entityCombinations = getAllCombinations(negatedFormatCombinations, negatedLanguageCombinations);

        List<TestCase> testCases = getCartesianProduct(searchTermCombinations, entityCombinations).stream().map(utterance -> {
            JsonObject expectedResponse = new JsonObject()
                .put(KEY_SEARCH_TERMS, SEARCH_TERM)
                .put(KEY_ENTITIES, new JsonArray()
                    .add(buildEntity(EntityType.FORMAT, FORMAT, true))
                    .add(buildEntity(EntityType.LANGUAGE, LANGUAGE, true)))
                .put(KEY_DATE_CONSTRAINTS, new JsonArray());

            return new TestCase(utterance, expectedResponse);
        }).collect(Collectors.toList());

        LOG.info("Testing format negated and language negated with [{}] utterances", testCases.size());
        handleTestResult(testRecursively(testCases), testContext);
    }

    @Test
    void testSpatialAndLanguage(VertxTestContext testContext) {
        List<String> searchTermCombinations = getIndicatorValueCombinations(SEARCH_TERM_INDICATOR, SEARCH_TERM);

        List<String> spatialCombinations = getIndicatorValueCombinations(SPATIAL_INDICATOR, SPATIAL);
        List<String> languageCombinations = getIndicatorValueCombinations(LANGUAGE_INDICATOR, LANGUAGE);
        List<String> entityCombinations = getCartesianProduct(spatialCombinations, languageCombinations);

        List<TestCase> testCases = getCartesianProduct(searchTermCombinations, entityCombinations).stream().map(utterance -> {
            JsonObject expectedResponse = new JsonObject()
                .put(KEY_SEARCH_TERMS, SEARCH_TERM)
                .put(KEY_ENTITIES, new JsonArray()
                    .add(buildEntity(EntityType.SPATIAL, SPATIAL, false))
                    .add(buildEntity(EntityType.LANGUAGE, LANGUAGE, false)))
                .put(KEY_DATE_CONSTRAINTS, new JsonArray());

            return new TestCase(utterance, expectedResponse);
        }).collect(Collectors.toList());

        LOG.info("Testing spatial and language with [{}] utterances", testCases.size());
        handleTestResult(testRecursively(testCases), testContext);
    }

    @Test
    void testSpatialNegatedAndLanguage(VertxTestContext testContext) {
        List<String> searchTermCombinations = getIndicatorValueCombinations(SEARCH_TERM_INDICATOR, SEARCH_TERM);

        List<String> negatedSpatialCombinations = getCartesianProduct(NEGATION, getIndicatorValueCombinations(SPATIAL_INDICATOR, SPATIAL));
        List<String> languageCombinations = getIndicatorValueCombinations(LANGUAGE_INDICATOR, LANGUAGE);
        List<String> entityCombinations = getCartesianProduct(negatedSpatialCombinations, languageCombinations);

        List<TestCase> testCases = getCartesianProduct(searchTermCombinations, entityCombinations).stream().map(utterance -> {
            JsonObject expectedResponse = new JsonObject()
                .put(KEY_SEARCH_TERMS, SEARCH_TERM)
                .put(KEY_ENTITIES, new JsonArray()
                    .add(buildEntity(EntityType.SPATIAL, SPATIAL, true))
                    .add(buildEntity(EntityType.LANGUAGE, LANGUAGE, false)))
                .put(KEY_DATE_CONSTRAINTS, new JsonArray());

            return new TestCase(utterance, expectedResponse);
        }).collect(Collectors.toList());

        LOG.info("Testing spatial negated and language with [{}] utterances", testCases.size());
        handleTestResult(testRecursively(testCases), testContext);
    }

    @Test
    void testSpatialAndLanguageNegated(VertxTestContext testContext) {
        List<String> searchTermCombinations = getIndicatorValueCombinations(SEARCH_TERM_INDICATOR, SEARCH_TERM);

        List<String> spatialCombinations = getIndicatorValueCombinations(SPATIAL_INDICATOR, SPATIAL);
        List<String> negatedLanguageCombinations = getCartesianProduct(NEGATION, getIndicatorValueCombinations(LANGUAGE_INDICATOR, LANGUAGE));
        List<String> entityCombinations = getCartesianProduct(spatialCombinations, negatedLanguageCombinations);

        List<TestCase> testCases = getCartesianProduct(searchTermCombinations, entityCombinations).stream().map(utterance -> {
            JsonObject expectedResponse = new JsonObject()
                .put(KEY_SEARCH_TERMS, SEARCH_TERM)
                .put(KEY_ENTITIES, new JsonArray()
                    .add(buildEntity(EntityType.SPATIAL, SPATIAL, false))
                    .add(buildEntity(EntityType.LANGUAGE, LANGUAGE, true)))
                .put(KEY_DATE_CONSTRAINTS, new JsonArray());

            return new TestCase(utterance, expectedResponse);
        }).collect(Collectors.toList());

        LOG.info("Testing spatial and language negated with [{}] utterances", testCases.size());
        handleTestResult(testRecursively(testCases), testContext);
    }

    @Test
    void testSpatialNegatedAndLanguageNegated(VertxTestContext testContext) {
        List<String> searchTermCombinations = getIndicatorValueCombinations(SEARCH_TERM_INDICATOR, SEARCH_TERM);

        List<String> negatedSpatialCombinations = getCartesianProduct(NEGATION, getIndicatorValueCombinations(SPATIAL_INDICATOR, SPATIAL));
        List<String> negatedLanguageCombinations = getCartesianProduct(NEGATION, getIndicatorValueCombinations(LANGUAGE_INDICATOR, LANGUAGE));
        List<String> entityCombinations = getCartesianProduct(negatedSpatialCombinations, negatedLanguageCombinations);

        List<TestCase> testCases = getCartesianProduct(searchTermCombinations, entityCombinations).stream().map(utterance -> {
            JsonObject expectedResponse = new JsonObject()
                .put(KEY_SEARCH_TERMS, SEARCH_TERM)
                .put(KEY_ENTITIES, new JsonArray()
                    .add(buildEntity(EntityType.SPATIAL, SPATIAL, true))
                    .add(buildEntity(EntityType.LANGUAGE, LANGUAGE, true)))
                .put(KEY_DATE_CONSTRAINTS, new JsonArray());

            return new TestCase(utterance, expectedResponse);
        }).collect(Collectors.toList());

        LOG.info("Testing spatial negated and language negated with [{}] utterances", testCases.size());
        handleTestResult(testRecursively(testCases), testContext);
    }

    @Test
    void testFormatAndSpatialAndLanguage(VertxTestContext testContext) {
        List<String> searchTermCombinations = getIndicatorValueCombinations(SEARCH_TERM_INDICATOR, SEARCH_TERM);
        List<String> formatCombinations = getIndicatorValueCombinations(FORMAT_INDICATOR, FORMAT);
        List<String> spatialCombinations = getIndicatorValueCombinations(SPATIAL_INDICATOR, SPATIAL);
        List<String> languageCombinations = getIndicatorValueCombinations(LANGUAGE_INDICATOR, LANGUAGE);

        List<String> entityCombinations = new ArrayList<>();

        // format,spatial,language & language,format,spatial
        entityCombinations.addAll(getAllCombinations(getCartesianProduct(formatCombinations, spatialCombinations), languageCombinations));
        // spatial,format,language & language,spatial,format
        entityCombinations.addAll(getAllCombinations(getCartesianProduct(spatialCombinations, formatCombinations), languageCombinations));
        // format,language,spatial
        entityCombinations.addAll(getCartesianProduct(getCartesianProduct(formatCombinations, languageCombinations), spatialCombinations));
        // spatial,language,format
        entityCombinations.addAll(getCartesianProduct(spatialCombinations, getCartesianProduct(languageCombinations, formatCombinations)));

        List<TestCase> testCases = getCartesianProduct(searchTermCombinations, entityCombinations).stream().map(utterance -> {
            JsonObject expectedResponse = new JsonObject()
                .put(KEY_SEARCH_TERMS, SEARCH_TERM)
                .put(KEY_ENTITIES, new JsonArray()
                    .add(buildEntity(EntityType.FORMAT, FORMAT, false))
                    .add(buildEntity(EntityType.SPATIAL, SPATIAL, false))
                    .add(buildEntity(EntityType.LANGUAGE, LANGUAGE, false)))
                .put(KEY_DATE_CONSTRAINTS, new JsonArray());

            return new TestCase(utterance, expectedResponse);
        }).collect(Collectors.toList());

        LOG.info("Testing format and spatial and language with [{}] utterances", testCases.size());
        handleTestResult(testRecursively(testCases), testContext);
    }

    @Test
    void testFormatNegatedAndSpatialAndLanguage(VertxTestContext testContext) {
        List<String> searchTermCombinations = getIndicatorValueCombinations(SEARCH_TERM_INDICATOR, SEARCH_TERM);
        List<String> negatedFormatCombinations = getCartesianProduct(NEGATION, getIndicatorValueCombinations(FORMAT_INDICATOR, FORMAT));
        List<String> spatialCombinations = getIndicatorValueCombinations(SPATIAL_INDICATOR, SPATIAL);
        List<String> languageCombinations = getIndicatorValueCombinations(LANGUAGE_INDICATOR, LANGUAGE);

        List<String> entityCombinations = new ArrayList<>();

        // format,spatial,language & language,format,spatial
        entityCombinations.addAll(getAllCombinations(getCartesianProduct(negatedFormatCombinations, spatialCombinations), languageCombinations));
        // spatial,format,language & language,spatial,format
        entityCombinations.addAll(getAllCombinations(getCartesianProduct(spatialCombinations, negatedFormatCombinations), languageCombinations));
        // format,language,spatial
        entityCombinations.addAll(getCartesianProduct(getCartesianProduct(negatedFormatCombinations, languageCombinations), spatialCombinations));
        // spatial,language,format
        entityCombinations.addAll(getCartesianProduct(spatialCombinations, getCartesianProduct(languageCombinations, negatedFormatCombinations)));

        List<TestCase> testCases = getCartesianProduct(searchTermCombinations, entityCombinations).stream().map(utterance -> {
            JsonObject expectedResponse = new JsonObject()
                .put(KEY_SEARCH_TERMS, SEARCH_TERM)
                .put(KEY_ENTITIES, new JsonArray()
                    .add(buildEntity(EntityType.FORMAT, FORMAT, true))
                    .add(buildEntity(EntityType.SPATIAL, SPATIAL, false))
                    .add(buildEntity(EntityType.LANGUAGE, LANGUAGE, false)))
                .put(KEY_DATE_CONSTRAINTS, new JsonArray());

            return new TestCase(utterance, expectedResponse);
        }).collect(Collectors.toList());

        LOG.info("Testing format negated and spatial and language with [{}] utterances", testCases.size());
        handleTestResult(testRecursively(testCases), testContext);
    }

    @Test
    void testFormatAndSpatialNegatedAndLanguage(VertxTestContext testContext) {
        List<String> searchTermCombinations = getIndicatorValueCombinations(SEARCH_TERM_INDICATOR, SEARCH_TERM);
        List<String> formatCombinations = getIndicatorValueCombinations(FORMAT_INDICATOR, FORMAT);
        List<String> negatedSpatialCombinations = getCartesianProduct(NEGATION, getIndicatorValueCombinations(SPATIAL_INDICATOR, SPATIAL));
        List<String> languageCombinations = getIndicatorValueCombinations(LANGUAGE_INDICATOR, LANGUAGE);

        List<String> entityCombinations = new ArrayList<>();

        // format,spatial,language & language,format,spatial
        entityCombinations.addAll(getAllCombinations(getCartesianProduct(formatCombinations, negatedSpatialCombinations), languageCombinations));
        // spatial,format,language & language,spatial,format
        entityCombinations.addAll(getAllCombinations(getCartesianProduct(negatedSpatialCombinations, formatCombinations), languageCombinations));
        // format,language,spatial
        entityCombinations.addAll(getCartesianProduct(getCartesianProduct(formatCombinations, languageCombinations), negatedSpatialCombinations));
        // spatial,language,format
        entityCombinations.addAll(getCartesianProduct(negatedSpatialCombinations, getCartesianProduct(languageCombinations, formatCombinations)));

        List<TestCase> testCases = getCartesianProduct(searchTermCombinations, entityCombinations).stream().map(utterance -> {
            JsonObject expectedResponse = new JsonObject()
                .put(KEY_SEARCH_TERMS, SEARCH_TERM)
                .put(KEY_ENTITIES, new JsonArray()
                    .add(buildEntity(EntityType.FORMAT, FORMAT, false))
                    .add(buildEntity(EntityType.SPATIAL, SPATIAL, true))
                    .add(buildEntity(EntityType.LANGUAGE, LANGUAGE, false)))
                .put(KEY_DATE_CONSTRAINTS, new JsonArray());

            return new TestCase(utterance, expectedResponse);
        }).collect(Collectors.toList());

        LOG.info("Testing format and spatial negated and language with [{}] utterances", testCases.size());
        handleTestResult(testRecursively(testCases), testContext);
    }

    @Test
    void testFormatAndSpatialAndLanguageNegated(VertxTestContext testContext) {
        List<String> searchTermCombinations = getIndicatorValueCombinations(SEARCH_TERM_INDICATOR, SEARCH_TERM);
        List<String> formatCombinations = getIndicatorValueCombinations(FORMAT_INDICATOR, FORMAT);
        List<String> spatialCombinations = getIndicatorValueCombinations(SPATIAL_INDICATOR, SPATIAL);
        List<String> negatedLanguageCombinations = getCartesianProduct(NEGATION, getIndicatorValueCombinations(LANGUAGE_INDICATOR, LANGUAGE));

        List<String> entityCombinations = new ArrayList<>();

        // format,spatial,language & language,format,spatial
        entityCombinations.addAll(getAllCombinations(getCartesianProduct(formatCombinations, spatialCombinations), negatedLanguageCombinations));
        // spatial,format,language & language,spatial,format
        entityCombinations.addAll(getAllCombinations(getCartesianProduct(spatialCombinations, formatCombinations), negatedLanguageCombinations));
        // format,language,spatial
        entityCombinations.addAll(getCartesianProduct(getCartesianProduct(formatCombinations, negatedLanguageCombinations), spatialCombinations));
        // spatial,language,format
        entityCombinations.addAll(getCartesianProduct(spatialCombinations, getCartesianProduct(negatedLanguageCombinations, formatCombinations)));

        List<TestCase> testCases = getCartesianProduct(searchTermCombinations, entityCombinations).stream().map(utterance -> {
            JsonObject expectedResponse = new JsonObject()
                .put(KEY_SEARCH_TERMS, SEARCH_TERM)
                .put(KEY_ENTITIES, new JsonArray()
                    .add(buildEntity(EntityType.FORMAT, FORMAT, false))
                    .add(buildEntity(EntityType.SPATIAL, SPATIAL, false))
                    .add(buildEntity(EntityType.LANGUAGE, LANGUAGE, true)))
                .put(KEY_DATE_CONSTRAINTS, new JsonArray());

            return new TestCase(utterance, expectedResponse);
        }).collect(Collectors.toList());

        LOG.info("Testing format and spatial and language negated with [{}] utterances", testCases.size());
        handleTestResult(testRecursively(testCases), testContext);
    }

    @Test
    void testFormatNegatedAndSpatialNegatedAndLanguage(VertxTestContext testContext) {
        List<String> searchTermCombinations = getIndicatorValueCombinations(SEARCH_TERM_INDICATOR, SEARCH_TERM);
        List<String> negatedFormatCombinations = getCartesianProduct(NEGATION, getIndicatorValueCombinations(FORMAT_INDICATOR, FORMAT));
        List<String> negatedSpatialCombinations = getCartesianProduct(NEGATION, getIndicatorValueCombinations(SPATIAL_INDICATOR, SPATIAL));
        List<String> languageCombinations = getIndicatorValueCombinations(LANGUAGE_INDICATOR, LANGUAGE);

        List<String> entityCombinations = new ArrayList<>();

        // format,spatial,language & language,format,spatial
        entityCombinations.addAll(getAllCombinations(getCartesianProduct(negatedFormatCombinations, negatedSpatialCombinations), languageCombinations));
        // spatial,format,language & language,spatial,format
        entityCombinations.addAll(getAllCombinations(getCartesianProduct(negatedSpatialCombinations, negatedFormatCombinations), languageCombinations));
        // format,language,spatial
        entityCombinations.addAll(getCartesianProduct(getCartesianProduct(negatedFormatCombinations, languageCombinations), negatedSpatialCombinations));
        // spatial,language,format
        entityCombinations.addAll(getCartesianProduct(negatedSpatialCombinations, getCartesianProduct(languageCombinations, negatedFormatCombinations)));

        List<TestCase> testCases = getCartesianProduct(searchTermCombinations, entityCombinations).stream().map(utterance -> {
            JsonObject expectedResponse = new JsonObject()
                .put(KEY_SEARCH_TERMS, SEARCH_TERM)
                .put(KEY_ENTITIES, new JsonArray()
                    .add(buildEntity(EntityType.FORMAT, FORMAT, true))
                    .add(buildEntity(EntityType.SPATIAL, SPATIAL, true))
                    .add(buildEntity(EntityType.LANGUAGE, LANGUAGE, false)))
                .put(KEY_DATE_CONSTRAINTS, new JsonArray());

            return new TestCase(utterance, expectedResponse);
        }).collect(Collectors.toList());

        LOG.info("Testing format negated and spatial negated and language with [{}] utterances", testCases.size());
        handleTestResult(testRecursively(testCases), testContext);
    }

    @Test
    void testFormatNegatedAndSpatialAndLanguageNegated(VertxTestContext testContext) {
        List<String> searchTermCombinations = getIndicatorValueCombinations(SEARCH_TERM_INDICATOR, SEARCH_TERM);
        List<String> negatedFormatCombinations = getCartesianProduct(NEGATION, getIndicatorValueCombinations(FORMAT_INDICATOR, FORMAT));
        List<String> spatialCombinations = getIndicatorValueCombinations(SPATIAL_INDICATOR, SPATIAL);
        List<String> negatedLanguageCombinations = getCartesianProduct(NEGATION, getIndicatorValueCombinations(LANGUAGE_INDICATOR, LANGUAGE));

        List<String> entityCombinations = new ArrayList<>();

        // format,spatial,language & language,format,spatial
        entityCombinations.addAll(getAllCombinations(getCartesianProduct(negatedFormatCombinations, spatialCombinations), negatedLanguageCombinations));
        // spatial,format,language & language,spatial,format
        entityCombinations.addAll(getAllCombinations(getCartesianProduct(spatialCombinations, negatedFormatCombinations), negatedLanguageCombinations));
        // format,language,spatial
        entityCombinations.addAll(getCartesianProduct(getCartesianProduct(negatedFormatCombinations, negatedLanguageCombinations), spatialCombinations));
        // spatial,language,format
        entityCombinations.addAll(getCartesianProduct(spatialCombinations, getCartesianProduct(negatedLanguageCombinations, negatedFormatCombinations)));

        List<TestCase> testCases = getCartesianProduct(searchTermCombinations, entityCombinations).stream().map(utterance -> {
            JsonObject expectedResponse = new JsonObject()
                .put(KEY_SEARCH_TERMS, SEARCH_TERM)
                .put(KEY_ENTITIES, new JsonArray()
                    .add(buildEntity(EntityType.FORMAT, FORMAT, true))
                    .add(buildEntity(EntityType.SPATIAL, SPATIAL, false))
                    .add(buildEntity(EntityType.LANGUAGE, LANGUAGE, true)))
                .put(KEY_DATE_CONSTRAINTS, new JsonArray());

            return new TestCase(utterance, expectedResponse);
        }).collect(Collectors.toList());

        LOG.info("Testing format negated and spatial and language negated with [{}] utterances", testCases.size());
        handleTestResult(testRecursively(testCases), testContext);
    }

    @Test
    void testFormatAndSpatialNegatedAndLanguageNegated(VertxTestContext testContext) {

        List<String> searchTermCombinations = getIndicatorValueCombinations(SEARCH_TERM_INDICATOR, SEARCH_TERM);
        List<String> formatCombinations = getIndicatorValueCombinations(FORMAT_INDICATOR, FORMAT);
        List<String> negatedSpatialCombinations = getCartesianProduct(NEGATION, getIndicatorValueCombinations(SPATIAL_INDICATOR, SPATIAL));
        List<String> negatedLanguageCombinations = getCartesianProduct(NEGATION, getIndicatorValueCombinations(LANGUAGE_INDICATOR, LANGUAGE));

        List<String> entityCombinations = new ArrayList<>();

        // format,spatial,language & language,format,spatial
        entityCombinations.addAll(getAllCombinations(getCartesianProduct(formatCombinations, negatedSpatialCombinations), negatedLanguageCombinations));
        // spatial,format,language & language,spatial,format
        entityCombinations.addAll(getAllCombinations(getCartesianProduct(negatedSpatialCombinations, formatCombinations), negatedLanguageCombinations));
        // format,language,spatial
        entityCombinations.addAll(getCartesianProduct(getCartesianProduct(formatCombinations, negatedLanguageCombinations), negatedSpatialCombinations));
        // spatial,language,format
        entityCombinations.addAll(getCartesianProduct(negatedSpatialCombinations, getCartesianProduct(negatedLanguageCombinations, formatCombinations)));

        List<TestCase> testCases = getCartesianProduct(searchTermCombinations, entityCombinations).stream().map(utterance -> {
            JsonObject expectedResponse = new JsonObject()
                .put(KEY_SEARCH_TERMS, SEARCH_TERM)
                .put(KEY_ENTITIES, new JsonArray()
                    .add(buildEntity(EntityType.FORMAT, FORMAT, false))
                    .add(buildEntity(EntityType.SPATIAL, SPATIAL, true))
                    .add(buildEntity(EntityType.LANGUAGE, LANGUAGE, true)))
                .put(KEY_DATE_CONSTRAINTS, new JsonArray());

            return new TestCase(utterance, expectedResponse);
        }).collect(Collectors.toList());

        LOG.info("Testing format and spatial negated and language negated with [{}] utterances", testCases.size());
        testContext.completeNow();
    }

    @Test
    void testFormatNegatedAndSpatialNegatedAndLanguageNegated(VertxTestContext testContext) {
        List<String> searchTermCombinations = getIndicatorValueCombinations(SEARCH_TERM_INDICATOR, SEARCH_TERM);
        List<String> negatedFormatCombinations = getCartesianProduct(NEGATION, getIndicatorValueCombinations(FORMAT_INDICATOR, FORMAT));
        List<String> negatedSpatialCombinations = getCartesianProduct(NEGATION, getIndicatorValueCombinations(SPATIAL_INDICATOR, SPATIAL));
        List<String> negatedLanguageCombinations = getCartesianProduct(NEGATION, getIndicatorValueCombinations(LANGUAGE_INDICATOR, LANGUAGE));

        List<String> entityCombinations = new ArrayList<>();

        // format,spatial,language & language,format,spatial
        entityCombinations.addAll(getAllCombinations(getCartesianProduct(negatedFormatCombinations, negatedSpatialCombinations), negatedLanguageCombinations));
        // spatial,format,language & language,spatial,format
        entityCombinations.addAll(getAllCombinations(getCartesianProduct(negatedSpatialCombinations, negatedFormatCombinations), negatedLanguageCombinations));
        // format,language,spatial
        entityCombinations.addAll(getCartesianProduct(getCartesianProduct(negatedFormatCombinations, negatedLanguageCombinations), negatedSpatialCombinations));
        // spatial,language,format
        entityCombinations.addAll(getCartesianProduct(negatedSpatialCombinations, getCartesianProduct(negatedLanguageCombinations, negatedFormatCombinations)));

        List<TestCase> testCases = getCartesianProduct(searchTermCombinations, entityCombinations).stream().map(utterance -> {
            JsonObject expectedResponse = new JsonObject()
                .put(KEY_SEARCH_TERMS, SEARCH_TERM)
                .put(KEY_ENTITIES, new JsonArray()
                    .add(buildEntity(EntityType.FORMAT, FORMAT, true))
                    .add(buildEntity(EntityType.SPATIAL, SPATIAL, true))
                    .add(buildEntity(EntityType.LANGUAGE, LANGUAGE, true)))
                .put(KEY_DATE_CONSTRAINTS, new JsonArray());

            return new TestCase(utterance, expectedResponse);
        }).collect(Collectors.toList());

        LOG.info("Testing format negated and spatial negated and language negated with [{}] utterances", testCases.size());
        handleTestResult(testRecursively(testCases), testContext);
    }

    @Test
    void testStartDate(VertxTestContext testContext) {
        List<String> searchTermCombinations = getIndicatorValueCombinations(SEARCH_TERM_INDICATOR, SEARCH_TERM);

        List<TestCase> testCases = new ArrayList<>();

        TEMPORAL_VALUES.forEach(temporal -> {
            List<String> utterances = getCartesianProduct(searchTermCombinations, getIndicatorValueCombinations(START_DATE_INDICATORS, temporal));

            JsonObject expectedResponse = new JsonObject()
                .put(KEY_SEARCH_TERMS, SEARCH_TERM)
                .put(KEY_ENTITIES, new JsonArray())
                .put(KEY_DATE_CONSTRAINTS, new JsonArray()
                    .add(new JsonObject().put("startDate", temporal)));

            testCases.addAll(utterances.stream().map(utterance -> new TestCase(utterance, expectedResponse)).collect(Collectors.toList()));
        });

        LOG.info("Testing start date with [{}] utterances", testCases.size());
        handleTestResult(testRecursively(testCases), testContext);
    }

    @Test
    void testEndDate(VertxTestContext testContext) {
        List<String> searchTermCombinations = getIndicatorValueCombinations(SEARCH_TERM_INDICATOR, SEARCH_TERM);

        List<TestCase> testCases = new ArrayList<>();

        TEMPORAL_VALUES.forEach(temporal -> {
            List<String> utterances = getCartesianProduct(searchTermCombinations, getIndicatorValueCombinations(END_DATE_INDICATORS, temporal));

            JsonObject expectedResponse = new JsonObject()
                .put(KEY_SEARCH_TERMS, SEARCH_TERM)
                .put(KEY_ENTITIES, new JsonArray())
                .put(KEY_DATE_CONSTRAINTS, new JsonArray()
                    .add(new JsonObject().put("endDate", temporal)));

            testCases.addAll(utterances.stream().map(utterance -> new TestCase(utterance, expectedResponse)).collect(Collectors.toList()));
        });

        LOG.info("Testing end date with [{}] utterances", testCases.size());
        handleTestResult(testRecursively(testCases), testContext);
    }

    @Test
    void testStartAndEndDate(VertxTestContext testContext) {
        List<String> searchTermCombinations = getIndicatorValueCombinations(SEARCH_TERM_INDICATOR, SEARCH_TERM);

        List<TestCase> testCases = new ArrayList<>();

        TEMPORAL_VALUES.forEach(startDate ->
            TEMPORAL_VALUES.forEach(endDate -> {
                List<String> utterances = getCartesianProduct(
                    getCartesianProduct(searchTermCombinations,
                        getIndicatorValueCombinations(START_DATE_INDICATORS, startDate)),
                    getIndicatorValueCombinations(END_DATE_INDICATORS, endDate)
                );

                JsonObject expectedResponse = new JsonObject()
                    .put(KEY_SEARCH_TERMS, SEARCH_TERM)
                    .put(KEY_ENTITIES, new JsonArray())
                    .put(KEY_DATE_CONSTRAINTS, new JsonArray()
                        .add(new JsonObject().put("startDate", startDate))
                        .add(new JsonObject().put("endDate", endDate)));


                testCases.addAll(utterances
                    .stream().map(utterance -> new TestCase(utterance, expectedResponse)).collect(Collectors.toList()));
            }));

        LOG.info("Testing end date with [{}] utterances", testCases.size());
        handleTestResult(testRecursively(testCases), testContext);
    }


    // Sends incoming requests to the ChatScript server
    private Future<Void> sendRequestToChatScript(String request) {
        Future<Void> future = Future.future();

        // ChatScript requires messages to consist of three values; user name, bot name, and message
        // If the bot's name is omitted, the default bot is used. All values must be separated by null bytes
        byte[] nullByte = {0};
        Buffer chatScriptRequest = Buffer.buffer("integrationTestUser".getBytes(StandardCharsets.UTF_8))
            .appendBytes(nullByte)
            .appendBytes(nullByte)
            .appendBytes(request.getBytes(StandardCharsets.UTF_8))
            .appendBytes(nullByte);

        netClient.connect(chatScriptPort, chatScriptHost, socket -> {
            if (socket.succeeded()) {
                socket.result().write(chatScriptRequest, writeHandler -> {
                    if (writeHandler.succeeded()) {
                        LOG.debug("Successfully sent request [{}] to ChatScript", request);
                        future.complete();
                    } else {
                        future.fail(writeHandler.cause());
                    }
                });
            } else {
                future.fail(socket.cause());
            }
        });

        return future;
    }

    // Recursively checks all test cases and creates a checkpoint for each
    // Futures from test are composed to prevent race conditions at the endpoint
    // This is intended to prevent timeouts for tests with a lot of test cases
    private Future<Void> testRecursively(List<TestCase> testCases) {
        testCase = testCases.get(0);
        testCompleted = Future.future();

        LOG.debug("Testing [{}], [{}] to go...", testCase.getUtterance().trim(), testCases.size() - 1);

        return testCases.size() > 1
            ? sendRequestToChatScript(":reset")
            .compose(resetHandler -> sendRequestToChatScript(testCase.getUtterance()))
            .compose(utteranceHandler -> testCompleted)
            .compose(testHandler ->
                testRecursively(testCases.subList(1, testCases.size())))
            : sendRequestToChatScript(":reset")
            .compose(resetHandler -> sendRequestToChatScript(testCase.getUtterance()))
            .compose(utteranceHandler -> testCompleted);
    }

    // Search terms and entities each have a list of possible phrases from which the ChatScript can deduct what is meant.
    // In order to test all these indicators they need to be combined with their actual value to be tested
    private List<String> getIndicatorValueCombinations(List<String> indicators, String value) {
        return indicators.stream().map(indicator -> indicator + " " + value).collect(Collectors.toList());
    }

    private List<String> getAllCombinations(List<String> listA, List<String> listB) {
        List<String> result = new ArrayList<>();
        result.addAll(getCartesianProduct(listA, listB));
        result.addAll(getCartesianProduct(listB, listA));
        return result;
    }

    // When multiple entities or a search term with entities is tested, all possible combinations need to be created.
    // This, i.e. the set of all possible combinations of two collections, is called the cartesian product.
    // Note that this method only calculates one order of combinations, e.g. {a, b} x {c} would yield {ac, bc} and not {ac, bc, ca, cb}
    private List<String> getCartesianProduct(List<String> listA, List<String> listB) {
        List<String> result = new ArrayList<>();

        listA.forEach(a -> listB.forEach(b -> result.add(a + " " + b)));

        return result;
    }

    // helper method that creates a JsonObject for later comparison
    private JsonObject buildEntity(EntityType type, String values, boolean negation) {
        return new JsonObject()
            .put("type", type.name().toUpperCase())
            .put("values", values)
            .put("negation", negation);
    }

    // helper method to reduce boilerplate code
    private void handleTestResult(Future<Void> future, VertxTestContext testContext) {
        future.setHandler(handler -> {
            if (handler.succeeded()) {
                testContext.completeNow();
            } else {
                testContext.failNow(handler.cause());
            }
        });
    }
}
